import re
from datetime import datetime, timedelta
from difflib import SequenceMatcher


class ParseError(ValueError):
    pass


TIME_UNITS = dict(
    seconds="seconds",
    minutes="minutes",
    hours="hours",
    days="days",
    weeks="weeks",
    months="months",
    years="years",
    # Shortcuts
    s="seconds",
    min="minutes",
    h="hours",
    d="days",
    w="weeks",
    mon="months",
    y="years",
)

EXTRAPOLATED = {"months": (30, "days"), "years": (365, "days")}
assert set(EXTRAPOLATED) <= set(TIME_UNITS)

TIME_RE = re.compile(r"(\d+)([a-z]+)")

UNITS_STR = ", ".join(sorted(TIME_UNITS.keys()))


def string_similarity(a, b):
    return SequenceMatcher(None, a, b).ratio()


def parse_time_atom(count, unit):
    count = int(count)
    try:
        unit = TIME_UNITS[unit]
    except KeyError:
        most_similar = max(TIME_UNITS, key=lambda k: string_similarity(k, unit))
        raise ParseError(
            f"'{unit}' is not a recognized time unit. Did you mean '{most_similar}'?" f"\nSupported units: {UNITS_STR}"
        )

    if unit in EXTRAPOLATED:
        mul, unit = EXTRAPOLATED[unit]
        count *= mul
    return count, unit


def parse_time_delta(t: str):
    time_dict = {}
    while t:
        m = TIME_RE.match(t)
        if not m:
            raise ParseError(f"Cannot parse '{t}': Not a recognized time delta")
        count, unit = parse_time_atom(*m.groups())
        if unit in time_dict:
            raise ParseError(f"Time unit {unit} specified more than once")
        time_dict[unit] = count
        t = t[m.end() :]

    if not time_dict:
        raise ParseError("No time difference specified")
    return timedelta(**time_dict)


def parse_time_before_now(t: str):
    return datetime.now() - parse_time_delta(t)



from .database_types import *
from .base import ThreadedDatabase, import_helper, ConnectError
from .base import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, _CHECKSUM_BITSIZE, TIMESTAMP_PRECISION_POS

SESSION_TIME_ZONE = None  # Changed by the tests


@import_helper("postgresql")
def import_postgresql():
    import psycopg2
    import psycopg2.extras

    psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)
    return psycopg2


class PostgreSQL(ThreadedDatabase):
    TYPE_CLASSES = {
        # Timestamps
        "timestamp with time zone": TimestampTZ,
        "timestamp without time zone": Timestamp,
        "timestamp": Timestamp,
        # Numbers
        "double precision": Float,
        "real": Float,
        "decimal": Decimal,
        "integer": Integer,
        "numeric": Decimal,
        "bigint": Integer,
        # Text
        "character": Text,
        "character varying": Text,
        "varchar": Text,
        "text": Text,
        # UUID
        "uuid": Native_UUID,
    }
    ROUNDS_ON_PREC_LOSS = True

    default_schema = "public"

    def __init__(self, *, thread_count, **kw):
        self._args = kw

        super().__init__(thread_count=thread_count)

    def _convert_db_precision_to_digits(self, p: int) -> int:
        # Subtracting 2 due to wierd precision issues in PostgreSQL
        return super()._convert_db_precision_to_digits(p) - 2

    def create_connection(self):
        if not self._args:
            self._args["host"] = None  # psycopg2 requires 1+ arguments

        pg = import_postgresql()
        try:
            c = pg.connect(**self._args)
            if SESSION_TIME_ZONE:
                c.cursor().execute(f"SET TIME ZONE '{SESSION_TIME_ZONE}'")
            return c
        except pg.OperationalError as e:
            raise ConnectError(*e.args) from e

    def quote(self, s: str):
        return f'"{s}"'

    def md5_to_int(self, s: str) -> str:
        return f"('x' || substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}))::bit({_CHECKSUM_BITSIZE})::bigint"

    def to_string(self, s: str):
        return f"{s}::varchar"

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            return f"to_char({value}::timestamp({coltype.precision}), 'YYYY-mm-dd HH24:MI:SS.US')"

        timestamp6 = f"to_char({value}::timestamp(6), 'YYYY-mm-dd HH24:MI:SS.US')"
        return (
            f"RPAD(LEFT({timestamp6}, {TIMESTAMP_PRECISION_POS+coltype.precision}), {TIMESTAMP_PRECISION_POS+6}, '0')"
        )

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return self.to_string(f"{value}::decimal(38, {coltype.precision})")



from .database_types import *
from .presto import Presto
from .base import import_helper
from .base import TIMESTAMP_PRECISION_POS


@import_helper("trino")
def import_trino():
    import trino

    return trino


class Trino(Presto):
    def __init__(self, **kw):
        trino = import_trino()

        self._conn = trino.dbapi.connect(**kw)

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            s = f"date_format(cast({value} as timestamp({coltype.precision})), '%Y-%m-%d %H:%i:%S.%f')"
        else:
            s = f"date_format(cast({value} as timestamp(6)), '%Y-%m-%d %H:%i:%S.%f')"

        return (
            f"RPAD(RPAD({s}, {TIMESTAMP_PRECISION_POS + coltype.precision}, '.'), {TIMESTAMP_PRECISION_POS + 6}, '0')"
        )

    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        return f"TRIM({value})"



from .database_types import *
from .base import ThreadedDatabase, import_helper, ConnectError
from .base import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, TIMESTAMP_PRECISION_POS


@import_helper("mysql")
def import_mysql():
    import mysql.connector

    return mysql.connector


class MySQL(ThreadedDatabase):
    TYPE_CLASSES = {
        # Dates
        "datetime": Datetime,
        "timestamp": Timestamp,
        # Numbers
        "double": Float,
        "float": Float,
        "decimal": Decimal,
        "int": Integer,
        "bigint": Integer,
        # Text
        "varchar": Text,
        "char": Text,
        "varbinary": Text,
        "binary": Text,
    }
    ROUNDS_ON_PREC_LOSS = True

    def __init__(self, *, thread_count, **kw):
        self._args = kw

        super().__init__(thread_count=thread_count)

        # In MySQL schema and database are synonymous
        self.default_schema = kw["database"]

    def create_connection(self):
        mysql = import_mysql()
        try:
            return mysql.connect(charset="utf8", use_unicode=True, **self._args)
        except mysql.Error as e:
            if e.errno == mysql.errorcode.ER_ACCESS_DENIED_ERROR:
                raise ConnectError("Bad user name or password") from e
            elif e.errno == mysql.errorcode.ER_BAD_DB_ERROR:
                raise ConnectError("Database does not exist") from e
            raise ConnectError(*e.args) from e

    def quote(self, s: str):
        return f"`{s}`"

    def md5_to_int(self, s: str) -> str:
        return f"cast(conv(substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16, 10) as unsigned)"

    def to_string(self, s: str):
        return f"cast({s} as char)"

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            return self.to_string(f"cast( cast({value} as datetime({coltype.precision})) as datetime(6))")

        s = self.to_string(f"cast({value} as datetime(6))")
        return f"RPAD(RPAD({s}, {TIMESTAMP_PRECISION_POS+coltype.precision}, '.'), {TIMESTAMP_PRECISION_POS+6}, '0')"

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return self.to_string(f"cast({value} as decimal(38, {coltype.precision}))")

    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        return f"TRIM(CAST({value} AS char))"



import math
import sys
import logging
from typing import Dict, Tuple, Optional, Sequence, Type, List
from functools import wraps
from concurrent.futures import ThreadPoolExecutor
import threading
from abc import abstractmethod

from data_diff.utils import is_uuid, safezip
from .database_types import (
    AbstractDatabase,
    ColType,
    Integer,
    Decimal,
    Float,
    ColType_UUID,
    Native_UUID,
    String_Alphanum,
    String_UUID,
    TemporalType,
    UnknownColType,
    Text,
    DbTime,
)
from data_diff.sql import DbPath, SqlOrStr, Compiler, Explain, Select, TableName

logger = logging.getLogger("database")


def parse_table_name(t):
    return tuple(t.split("."))


def import_helper(package: str = None, text=""):
    def dec(f):
        @wraps(f)
        def _inner():
            try:
                return f()
            except ModuleNotFoundError as e:
                s = text
                if package:
                    s += f"You can install it using 'pip install data-diff[{package}]'."
                raise ModuleNotFoundError(f"{e}\n\n{s}\n")

        return _inner

    return dec


class ConnectError(Exception):
    pass


class QueryError(Exception):
    pass


def _one(seq):
    (x,) = seq
    return x


def _query_conn(conn, sql_code: str) -> list:
    c = conn.cursor()
    c.execute(sql_code)
    if sql_code.lower().startswith("select"):
        return c.fetchall()


class Database(AbstractDatabase):
    """Base abstract class for databases.

    Used for providing connection code and implementation specific SQL utilities.

    Instanciated using :meth:`~data_diff.connect`
    """

    TYPE_CLASSES: Dict[str, type] = {}
    default_schema: str = None

    @property
    def name(self):
        return type(self).__name__

    def query(self, sql_ast: SqlOrStr, res_type: type):
        "Query the given SQL code/AST, and attempt to convert the result to type 'res_type'"

        compiler = Compiler(self)
        sql_code = compiler.compile(sql_ast)
        logger.debug("Running SQL (%s): %s", type(self).__name__, sql_code)
        if getattr(self, "_interactive", False) and isinstance(sql_ast, Select):
            explained_sql = compiler.compile(Explain(sql_ast))
            logger.info("EXPLAIN for SQL SELECT")
            logger.info(self._query(explained_sql))
            answer = input("Continue? [y/n] ")
            if not answer.lower() in ["y", "yes"]:
                sys.exit(1)

        res = self._query(sql_code)
        if res_type is int:
            res = _one(_one(res))
            if res is None:  # May happen due to sum() of 0 items
                return None
            return int(res)
        elif res_type is tuple:
            assert len(res) == 1, (sql_code, res)
            return res[0]
        elif getattr(res_type, "__origin__", None) is list and len(res_type.__args__) == 1:
            if res_type.__args__ in ((int,), (str,)):
                return [_one(row) for row in res]
            elif res_type.__args__ == (Tuple,):
                return [tuple(row) for row in res]
            else:
                raise ValueError(res_type)
        return res

    def enable_interactive(self):
        self._interactive = True

    def _convert_db_precision_to_digits(self, p: int) -> int:
        """Convert from binary precision, used by floats, to decimal precision."""
        # See: https://en.wikipedia.org/wiki/Single-precision_floating-point_format
        return math.floor(math.log(2**p, 10))

    def _parse_type_repr(self, type_repr: str) -> Optional[Type[ColType]]:
        return self.TYPE_CLASSES.get(type_repr)

    def _parse_type(
        self,
        table_path: DbPath,
        col_name: str,
        type_repr: str,
        datetime_precision: int = None,
        numeric_precision: int = None,
        numeric_scale: int = None,
    ) -> ColType:
        """ """

        cls = self._parse_type_repr(type_repr)
        if not cls:
            return UnknownColType(type_repr)

        if issubclass(cls, TemporalType):
            return cls(
                precision=datetime_precision if datetime_precision is not None else DEFAULT_DATETIME_PRECISION,
                rounds=self.ROUNDS_ON_PREC_LOSS,
            )

        elif issubclass(cls, Integer):
            return cls()

        elif issubclass(cls, Decimal):
            if numeric_scale is None:
                numeric_scale = 0  # Needed for Oracle.
            return cls(precision=numeric_scale)

        elif issubclass(cls, Float):
            # assert numeric_scale is None
            return cls(
                precision=self._convert_db_precision_to_digits(
                    numeric_precision if numeric_precision is not None else DEFAULT_NUMERIC_PRECISION
                )
            )

        elif issubclass(cls, (Text, Native_UUID)):
            return cls()

        raise TypeError(f"Parsing {type_repr} returned an unknown type '{cls}'.")

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)

        return (
            "SELECT column_name, data_type, datetime_precision, numeric_precision, numeric_scale "
            "FROM information_schema.columns "
            f"WHERE table_name = '{table}' AND table_schema = '{schema}'"
        )

    def query_table_schema(self, path: DbPath) -> Dict[str, tuple]:
        rows = self.query(self.select_table_schema(path), list)
        if not rows:
            raise RuntimeError(f"{self.name}: Table '{'.'.join(path)}' does not exist, or has no columns")

        d = {r[0]: r for r in rows}
        assert len(d) == len(rows)
        return d

    def _process_table_schema(self, path: DbPath, raw_schema: Dict[str, tuple], filter_columns: Sequence[str]):
        accept = {i.lower() for i in filter_columns}

        col_dict = {row[0]: self._parse_type(path, *row) for name, row in raw_schema.items() if name.lower() in accept}

        self._refine_coltypes(path, col_dict)

        # Return a dict of form {name: type} after normalization
        return col_dict

    def _refine_coltypes(self, table_path: DbPath, col_dict: Dict[str, ColType]):
        "Refine the types in the column dict, by querying the database for a sample of their values"

        text_columns = [k for k, v in col_dict.items() if isinstance(v, Text)]
        if not text_columns:
            return

        fields = [self.normalize_uuid(c, String_UUID()) for c in text_columns]
        samples_by_row = self.query(Select(fields, TableName(table_path), limit=16), list)
        if not samples_by_row:
            raise ValueError(f"Table {table_path} is empty.")

        samples_by_col = list(zip(*samples_by_row))

        for col_name, samples in safezip(text_columns, samples_by_col):
            uuid_samples = [s for s in samples if s and is_uuid(s)]

            if uuid_samples:
                if len(uuid_samples) != len(samples):
                    logger.warning(
                        f"Mixed UUID/Non-UUID values detected in column {'.'.join(table_path)}.{col_name}, disabling UUID support."
                    )
                else:
                    assert col_name in col_dict
                    col_dict[col_name] = String_UUID()
                    continue

            alphanum_samples = [s for s in samples if s and String_Alphanum.test_value(s)]
            if alphanum_samples:
                if len(alphanum_samples) != len(samples):
                    logger.warning(
                        f"Mixed Alphanum/Non-Alphanum values detected in column {'.'.join(table_path)}.{col_name}, disabling Alphanum support."
                    )
                else:
                    assert col_name in col_dict
                    lens = set(map(len, alphanum_samples))
                    if len(lens) > 1:
                        logger.warning(
                            f"Mixed Alphanum lengths detected in column {'.'.join(table_path)}.{col_name}, disabling Alphanum support."
                        )
                    else:
                        (length,) = lens
                        col_dict[col_name] = String_Alphanum(length=length)
                        continue

    # @lru_cache()
    # def get_table_schema(self, path: DbPath) -> Dict[str, ColType]:
    #     return self.query_table_schema(path)

    def _normalize_table_path(self, path: DbPath) -> DbPath:
        if len(path) == 1:
            if self.default_schema:
                return self.default_schema, path[0]
        elif len(path) != 2:
            raise ValueError(f"{self.name}: Bad table path for {self}: '{'.'.join(path)}'. Expected form: schema.table")

        return path

    def parse_table_name(self, name: str) -> DbPath:
        return parse_table_name(name)

    def offset_limit(self, offset: Optional[int] = None, limit: Optional[int] = None):
        if offset:
            raise NotImplementedError("No support for OFFSET in query")

        return f"LIMIT {limit}"

    def concat(self, l: List[str]) -> str:
        joined_exprs = ", ".join(l)
        return f"concat({joined_exprs})"

    def timestamp_value(self, t: DbTime) -> str:
        return f"'{t.isoformat()}'"

    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        if isinstance(coltype, String_UUID):
            return f"TRIM({value})"
        return self.to_string(value)


class ThreadedDatabase(Database):
    """Access the database through singleton threads.

    Used for database connectors that do not support sharing their connection between different threads.
    """

    def __init__(self, thread_count=1):
        self._init_error = None
        self._queue = ThreadPoolExecutor(thread_count, initializer=self.set_conn)
        self.thread_local = threading.local()
        logger.info(f"[{self.name}] Starting a threadpool, size={thread_count}.")

    def set_conn(self):
        assert not hasattr(self.thread_local, "conn")
        try:
            self.thread_local.conn = self.create_connection()
        except ModuleNotFoundError as e:
            self._init_error = e

    def _query(self, sql_code: str):
        r = self._queue.submit(self._query_in_worker, sql_code)
        return r.result()

    def _query_in_worker(self, sql_code: str):
        "This method runs in a worker thread"
        if self._init_error:
            raise self._init_error
        return _query_conn(self.thread_local.conn, sql_code)

    @abstractmethod
    def create_connection(self):
        ...

    def close(self):
        self._queue.shutdown()


CHECKSUM_HEXDIGITS = 15  # Must be 15 or lower
MD5_HEXDIGITS = 32

_CHECKSUM_BITSIZE = CHECKSUM_HEXDIGITS << 2
CHECKSUM_MASK = (2**_CHECKSUM_BITSIZE) - 1

DEFAULT_DATETIME_PRECISION = 6
DEFAULT_NUMERIC_PRECISION = 24

TIMESTAMP_PRECISION_POS = 20  # len("2022-06-03 12:24:35.") == 20



import re

from ..utils import match_regexps

from .database_types import *
from .base import Database, import_helper
from .base import (
    MD5_HEXDIGITS,
    CHECKSUM_HEXDIGITS,
    TIMESTAMP_PRECISION_POS,
)


@import_helper("presto")
def import_presto():
    import prestodb

    return prestodb


class Presto(Database):
    default_schema = "public"
    TYPE_CLASSES = {
        # Timestamps
        "timestamp with time zone": TimestampTZ,
        "timestamp without time zone": Timestamp,
        "timestamp": Timestamp,
        # Numbers
        "integer": Integer,
        "bigint": Integer,
        "real": Float,
        "double": Float,
        # Text
        "varchar": Text,
    }
    ROUNDS_ON_PREC_LOSS = True

    def __init__(self, **kw):
        prestodb = import_presto()

        if kw.get("schema"):
            self.default_schema = kw.get("schema")

        if kw.get("auth") == "basic":  # if auth=basic, add basic authenticator for Presto
            kw["auth"] = prestodb.auth.BasicAuthentication(kw.pop("user"), kw.pop("password"))

        if "cert" in kw:  # if a certificate was specified in URI, verify session with cert
            cert = kw.pop("cert")
            self._conn = prestodb.dbapi.connect(**kw)
            self._conn._http_session.verify = cert
        else:
            self._conn = prestodb.dbapi.connect(**kw)

    def quote(self, s: str):
        return f'"{s}"'

    def md5_to_int(self, s: str) -> str:
        return f"cast(from_base(substr(to_hex(md5(to_utf8({s}))), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16) as decimal(38, 0))"

    def to_string(self, s: str):
        return f"cast({s} as varchar)"

    def _query(self, sql_code: str) -> list:
        "Uses the standard SQL cursor interface"
        c = self._conn.cursor()
        c.execute(sql_code)
        if sql_code.lower().startswith("select"):
            return c.fetchall()
        # Required for the query to actually run 🤯
        if re.match(r"(insert|create|truncate|drop)", sql_code, re.IGNORECASE):
            return c.fetchone()

    def close(self):
        self._conn.close()

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        # TODO
        if coltype.rounds:
            s = f"date_format(cast({value} as timestamp(6)), '%Y-%m-%d %H:%i:%S.%f')"
        else:
            s = f"date_format(cast({value} as timestamp(6)), '%Y-%m-%d %H:%i:%S.%f')"

        return f"RPAD(RPAD({s}, {TIMESTAMP_PRECISION_POS+coltype.precision}, '.'), {TIMESTAMP_PRECISION_POS+6}, '0')"

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return self.to_string(f"cast({value} as decimal(38,{coltype.precision}))")

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)

        return (
            "SELECT column_name, data_type, 3 as datetime_precision, 3 as numeric_precision "
            "FROM INFORMATION_SCHEMA.COLUMNS "
            f"WHERE table_name = '{table}' AND table_schema = '{schema}'"
        )

    def _parse_type(
        self,
        table_path: DbPath,
        col_name: str,
        type_repr: str,
        datetime_precision: int = None,
        numeric_precision: int = None,
    ) -> ColType:
        timestamp_regexps = {
            r"timestamp\((\d)\)": Timestamp,
            r"timestamp\((\d)\) with time zone": TimestampTZ,
        }
        for m, t_cls in match_regexps(timestamp_regexps, type_repr):
            precision = int(m.group(1))
            return t_cls(precision=precision, rounds=self.ROUNDS_ON_PREC_LOSS)

        number_regexps = {r"decimal\((\d+),(\d+)\)": Decimal}
        for m, n_cls in match_regexps(number_regexps, type_repr):
            _prec, scale = map(int, m.groups())
            return n_cls(scale)

        string_regexps = {r"varchar\((\d+)\)": Text, r"char\((\d+)\)": Text}
        for m, n_cls in match_regexps(string_regexps, type_repr):
            return n_cls()

        return super()._parse_type(table_path, col_name, type_repr, datetime_precision, numeric_precision)

    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        # Trim doesn't work on CHAR type
        return f"TRIM(CAST({value} AS VARCHAR))"



# class MsSQL(ThreadedDatabase):
#     "AKA sql-server"

#     def __init__(self, host, port, user, password, *, database, thread_count, **kw):
#         args = dict(server=host, port=port, database=database, user=user, password=password, **kw)
#         self._args = {k: v for k, v in args.items() if v is not None}

#         super().__init__(thread_count=thread_count)

#     def create_connection(self):
#         mssql = import_mssql()
#         try:
#             return mssql.connect(**self._args)
#         except mssql.Error as e:
#             raise ConnectError(*e.args) from e

#     def quote(self, s: str):
#         return f"[{s}]"

#     def md5_to_int(self, s: str) -> str:
#         return f"CONVERT(decimal(38,0), CONVERT(bigint, HashBytes('MD5', {s}), 2))"
#         # return f"CONVERT(bigint, (CHECKSUM({s})))"

#     def to_string(self, s: str):
#         return f"CONVERT(varchar, {s})"



from typing import Tuple, Iterator, Optional, Union

from .tracking import disable_tracking
from .databases.connect import connect
from .diff_tables import (
    TableSegment,
    TableDiffer,
    DEFAULT_BISECTION_THRESHOLD,
    DEFAULT_BISECTION_FACTOR,
    DbKey,
    DbTime,
    DbPath,
)


def connect_to_table(
    db_info: Union[str, dict],
    table_name: Union[DbPath, str],
    key_column: str = "id",
    thread_count: Optional[int] = 1,
    **kwargs,
):
    """Connects to the given database, and creates a TableSegment instance

    Parameters:
        db_info: Either a URI string, or a dict of connection options.
        table_name: Name of the table as a string, or a tuple that signifies the path.
        key_column: Name of the key column
        thread_count: Number of threads for this connection (only if using a threadpooled implementation)
    """

    db = connect(db_info, thread_count=thread_count)

    if isinstance(table_name, str):
        table_name = db.parse_table_name(table_name)

    return TableSegment(db, table_name, key_column, **kwargs)


def diff_tables(
    table1: TableSegment,
    table2: TableSegment,
    *,
    # Name of the key column, which uniquely identifies each row (usually id)
    key_column: str = None,
    # Name of updated column, which signals that rows changed (usually updated_at or last_update)
    update_column: str = None,
    # Extra columns to compare
    extra_columns: Tuple[str, ...] = None,
    # Start/end key_column values, used to restrict the segment
    min_key: DbKey = None,
    max_key: DbKey = None,
    # Start/end update_column values, used to restrict the segment
    min_update: DbTime = None,
    max_update: DbTime = None,
    # Into how many segments to bisect per iteration
    bisection_factor: int = DEFAULT_BISECTION_FACTOR,
    # When should we stop bisecting and compare locally (in row count)
    bisection_threshold: int = DEFAULT_BISECTION_THRESHOLD,
    # Enable/disable threaded diffing. Needed to take advantage of database threads.
    threaded: bool = True,
    # Maximum size of each threadpool. None = auto. Only relevant when threaded is True.
    # There may be many pools, so number of actual threads can be a lot higher.
    max_threadpool_size: Optional[int] = 1,
    # Enable/disable debug prints
    debug: bool = False,
) -> Iterator:
    """Efficiently finds the diff between table1 and table2.

    Example:
        >>> table1 = connect_to_table('postgresql:///', 'Rating', 'id')
        >>> list(diff_tables(table1, table1))
        []

    """
    tables = [table1, table2]
    override_attrs = {
        k: v
        for k, v in dict(
            key_column=key_column,
            update_column=update_column,
            extra_columns=extra_columns,
            min_key=min_key,
            max_key=max_key,
            min_update=min_update,
            max_update=max_update,
        ).items()
        if v is not None
    }

    segments = [t.new(**override_attrs) for t in tables] if override_attrs else tables

    differ = TableDiffer(
        bisection_factor=bisection_factor,
        bisection_threshold=bisection_threshold,
        debug=debug,
        threaded=threaded,
        max_threadpool_size=max_threadpool_size,
    )
    return differ.diff_tables(*segments)



from typing import Type, List, Optional, Union
from itertools import zip_longest
import dsnparse

from runtype import dataclass

from .base import Database, ThreadedDatabase
from .postgresql import PostgreSQL
from .mysql import MySQL
from .oracle import Oracle
from .snowflake import Snowflake
from .bigquery import BigQuery
from .redshift import Redshift
from .presto import Presto
from .databricks import Databricks
from .trino import Trino


@dataclass
class MatchUriPath:
    database_cls: Type[Database]
    params: List[str]
    kwparams: List[str] = []
    help_str: str

    def match_path(self, dsn):
        dsn_dict = dict(dsn.query)
        matches = {}
        for param, arg in zip_longest(self.params, dsn.paths):
            if param is None:
                raise ValueError(f"Too many parts to path. Expected format: {self.help_str}")

            optional = param.endswith("?")
            param = param.rstrip("?")

            if arg is None:
                try:
                    arg = dsn_dict.pop(param)
                except KeyError:
                    if not optional:
                        raise ValueError(f"URI must specify '{param}'. Expected format: {self.help_str}")

                    arg = None

            assert param and param not in matches
            matches[param] = arg

        for param in self.kwparams:
            try:
                arg = dsn_dict.pop(param)
            except KeyError:
                raise ValueError(f"URI must specify '{param}'. Expected format: {self.help_str}")

            assert param and arg and param not in matches, (param, arg, matches.keys())
            matches[param] = arg

        for param, value in dsn_dict.items():
            if param in matches:
                raise ValueError(
                    f"Parameter '{param}' already provided as positional argument. Expected format: {self.help_str}"
                )

            matches[param] = value

        return matches


MATCH_URI_PATH = {
    "postgresql": MatchUriPath(PostgreSQL, ["database?"], help_str="postgresql://<user>:<pass>@<host>/<database>"),
    "mysql": MatchUriPath(MySQL, ["database?"], help_str="mysql://<user>:<pass>@<host>/<database>"),
    "oracle": MatchUriPath(Oracle, ["database?"], help_str="oracle://<user>:<pass>@<host>/<database>"),
    # "mssql": MatchUriPath(MsSQL, ["database?"], help_str="mssql://<user>:<pass>@<host>/<database>"),
    "redshift": MatchUriPath(Redshift, ["database?"], help_str="redshift://<user>:<pass>@<host>/<database>"),
    "snowflake": MatchUriPath(
        Snowflake,
        ["database", "schema"],
        ["warehouse"],
        help_str="snowflake://<user>:<pass>@<account>/<database>/<SCHEMA>?warehouse=<WAREHOUSE>",
    ),
    "presto": MatchUriPath(Presto, ["catalog", "schema"], help_str="presto://<user>@<host>/<catalog>/<schema>"),
    "bigquery": MatchUriPath(BigQuery, ["dataset"], help_str="bigquery://<project>/<dataset>"),
    "databricks": MatchUriPath(
        Databricks,
        ["catalog", "schema"],
        help_str="databricks://:access_token@server_name/http_path",
    ),
    "trino": MatchUriPath(Trino, ["catalog", "schema"], help_str="trino://<user>@<host>/<catalog>/<schema>"),
}


def connect_to_uri(db_uri: str, thread_count: Optional[int] = 1) -> Database:
    """Connect to the given database uri

    thread_count determines the max number of worker threads per database,
    if relevant. None means no limit.

    Parameters:
        db_uri (str): The URI for the database to connect
        thread_count (int, optional): Size of the threadpool. Ignored by cloud databases. (default: 1)

    Note: For non-cloud databases, a low thread-pool size may be a performance bottleneck.

    Supported schemes:
    - postgresql
    - mysql
    - oracle
    - snowflake
    - bigquery
    - redshift
    - presto
    - databricks
    - trino
    """

    dsn = dsnparse.parse(db_uri)
    if len(dsn.schemes) > 1:
        raise NotImplementedError("No support for multiple schemes")
    (scheme,) = dsn.schemes

    try:
        matcher = MATCH_URI_PATH[scheme]
    except KeyError:
        raise NotImplementedError(f"Scheme {scheme} currently not supported")

    cls = matcher.database_cls

    if scheme == "databricks":
        assert not dsn.user
        kw = {}
        kw["access_token"] = dsn.password
        kw["http_path"] = dsn.path
        kw["server_hostname"] = dsn.host
        kw.update(dsn.query)
    else:
        kw = matcher.match_path(dsn)

        if scheme == "bigquery":
            kw["project"] = dsn.host
            return cls(**kw)

        if scheme == "snowflake":
            kw["account"] = dsn.host
            assert not dsn.port
            kw["user"] = dsn.user
            kw["password"] = dsn.password
        else:
            kw["host"] = dsn.host
            kw["port"] = dsn.port
            kw["user"] = dsn.user
            if dsn.password:
                kw["password"] = dsn.password

    kw = {k: v for k, v in kw.items() if v is not None}

    if issubclass(cls, ThreadedDatabase):
        return cls(thread_count=thread_count, **kw)

    return cls(**kw)


def connect_with_dict(d, thread_count):
    d = dict(d)
    driver = d.pop("driver")
    try:
        matcher = MATCH_URI_PATH[driver]
    except KeyError:
        raise NotImplementedError(f"Driver {driver} currently not supported")

    cls = matcher.database_cls
    if issubclass(cls, ThreadedDatabase):
        return cls(thread_count=thread_count, **d)

    return cls(**d)


def connect(db_conf: Union[str, dict], thread_count: Optional[int] = 1) -> Database:
    """Connect to a database using the given database configuration.

    Configuration can be given either as a URI string, or as a dict of {option: value}.

    thread_count determines the max number of worker threads per database,
    if relevant. None means no limit.

    Parameters:
        db_conf (str | dict): The configuration for the database to connect. URI or dict.
        thread_count (int, optional): Size of the threadpool. Ignored by cloud databases. (default: 1)

    Note: For non-cloud databases, a low thread-pool size may be a performance bottleneck.

    Supported drivers:
    - postgresql
    - mysql
    - oracle
    - snowflake
    - bigquery
    - redshift
    - presto
    - databricks
    - trino
    """
    if isinstance(db_conf, str):
        return connect_to_uri(db_conf, thread_count)
    elif isinstance(db_conf, dict):
        return connect_with_dict(db_conf, thread_count)
    raise TypeError(db_conf)



from .database_types import *
from .base import Database, import_helper, parse_table_name, ConnectError
from .base import TIMESTAMP_PRECISION_POS


@import_helper(text="Please install BigQuery and configure your google-cloud access.")
def import_bigquery():
    from google.cloud import bigquery

    return bigquery


class BigQuery(Database):
    TYPE_CLASSES = {
        # Dates
        "TIMESTAMP": Timestamp,
        "DATETIME": Datetime,
        # Numbers
        "INT64": Integer,
        "INT32": Integer,
        "NUMERIC": Decimal,
        "BIGNUMERIC": Decimal,
        "FLOAT64": Float,
        "FLOAT32": Float,
        # Text
        "STRING": Text,
    }
    ROUNDS_ON_PREC_LOSS = False  # Technically BigQuery doesn't allow implicit rounding or truncation

    def __init__(self, project, *, dataset, **kw):
        bigquery = import_bigquery()

        self._client = bigquery.Client(project, **kw)
        self.project = project
        self.dataset = dataset

        self.default_schema = dataset

    def quote(self, s: str):
        return f"`{s}`"

    def md5_to_int(self, s: str) -> str:
        return f"cast(cast( ('0x' || substr(TO_HEX(md5({s})), 18)) as int64) as numeric)"

    def _normalize_returned_value(self, value):
        if isinstance(value, bytes):
            return value.decode()
        return value

    def _query(self, sql_code: str):
        from google.cloud import bigquery

        try:
            res = list(self._client.query(sql_code))
        except Exception as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise ConnectError(msg % (sql_code, e))

        if res and isinstance(res[0], bigquery.table.Row):
            res = [tuple(self._normalize_returned_value(v) for v in row.values()) for row in res]
        return res

    def to_string(self, s: str):
        return f"cast({s} as string)"

    def close(self):
        self._client.close()

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)

        return (
            f"SELECT column_name, data_type, 6 as datetime_precision, 38 as numeric_precision, 9 as numeric_scale FROM {schema}.INFORMATION_SCHEMA.COLUMNS "
            f"WHERE table_name = '{table}' AND table_schema = '{schema}'"
        )

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            timestamp = f"timestamp_micros(cast(round(unix_micros(cast({value} as timestamp))/1000000, {coltype.precision})*1000000 as int))"
            return f"FORMAT_TIMESTAMP('%F %H:%M:%E6S', {timestamp})"

        if coltype.precision == 0:
            return f"FORMAT_TIMESTAMP('%F %H:%M:%S.000000, {value})"
        elif coltype.precision == 6:
            return f"FORMAT_TIMESTAMP('%F %H:%M:%E6S', {value})"

        timestamp6 = f"FORMAT_TIMESTAMP('%F %H:%M:%E6S', {value})"
        return (
            f"RPAD(LEFT({timestamp6}, {TIMESTAMP_PRECISION_POS+coltype.precision}), {TIMESTAMP_PRECISION_POS+6}, '0')"
        )

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return f"format('%.{coltype.precision}f', {value})"

    def parse_table_name(self, name: str) -> DbPath:
        path = parse_table_name(name)
        return self._normalize_table_path(path)



from .base import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, QueryError, ConnectError

from .postgresql import PostgreSQL
from .mysql import MySQL
from .oracle import Oracle
from .snowflake import Snowflake
from .bigquery import BigQuery
from .redshift import Redshift
from .presto import Presto
from .databricks import Databricks
from .trino import Trino

from .connect import connect_to_uri



from typing import Any, Dict
import toml


class ConfigParseError(Exception):
    pass


def is_uri(s: str) -> bool:
    return "://" in s


def _apply_config(config: Dict[str, Any], run_name: str, kw: Dict[str, Any]):
    # Load config
    databases = config.pop("database", {})
    runs = config.pop("run", {})
    if config:
        raise ConfigParseError(f"Unknown option(s): {config}")

    # Init run_args
    run_args = runs.get("default") or {}
    if run_name:
        if run_name not in runs:
            raise ConfigParseError(f"Cannot find run '{run_name}' in configuration.")
        run_args.update(runs[run_name])
    else:
        run_name = "default"

    # Process databases + tables
    for index in "12":
        args = run_args.pop(index, {})
        for attr in ("database", "table"):
            if attr not in args:
                raise ConfigParseError(f"Running 'run.{run_name}': Connection #{index} in missing attribute '{attr}'.")

        database = args.pop("database")
        table = args.pop("table")
        threads = args.pop("threads", None)
        if args:
            raise ConfigParseError(f"Unexpected attributes for connection #{index}: {args}")

        if not is_uri(database):
            if database not in databases:
                raise ConfigParseError(
                    f"Database '{database}' not found in list of databases. Available: {list(databases)}."
                )
            database = dict(databases[database])
            assert isinstance(database, dict)
            if "driver" not in database:
                raise ConfigParseError(f"Database '{database}' did not specify a driver.")

        run_args[f"database{index}"] = database
        run_args[f"table{index}"] = table
        if threads is not None:
            run_args[f"threads{index}"] = int(threads)

    # Update keywords
    new_kw = dict(kw)  # Set defaults
    new_kw.update(run_args)  # Apply config
    new_kw.update({k: v for k, v in kw.items() if v})  # Apply non-empty defaults

    new_kw["__conf__"] = run_args

    return new_kw


def apply_config_from_file(path: str, run_name: str, kw: Dict[str, Any]):
    with open(path) as f:
        return _apply_config(toml.load(f), run_name, kw)


def apply_config_from_string(toml_config: str, run_name: str, kw: Dict[str, Any]):
    return _apply_config(toml.loads(toml_config), run_name, kw)



import decimal
from abc import ABC, abstractmethod
from typing import Sequence, Optional, Tuple, Union, Dict, List
from datetime import datetime

from runtype import dataclass

from data_diff.utils import ArithAlphanumeric, ArithUUID, CaseAwareMapping


DbPath = Tuple[str, ...]
DbKey = Union[int, str, bytes, ArithUUID, ArithAlphanumeric]
DbTime = datetime


class ColType:
    supported = True


@dataclass
class PrecisionType(ColType):
    precision: int
    rounds: bool


class TemporalType(PrecisionType):
    pass


class Timestamp(TemporalType):
    pass


class TimestampTZ(TemporalType):
    pass


class Datetime(TemporalType):
    pass


@dataclass
class NumericType(ColType):
    # 'precision' signifies how many fractional digits (after the dot) we want to compare
    precision: int


class FractionalType(NumericType):
    pass


class Float(FractionalType):
    pass


class IKey(ABC):
    "Interface for ColType, for using a column as a key in data-diff"
    python_type: type

    def make_value(self, value):
        return self.python_type(value)


class Decimal(FractionalType, IKey):  # Snowflake may use Decimal as a key
    @property
    def python_type(self) -> type:
        if self.precision == 0:
            return int
        return decimal.Decimal


class StringType(ColType):
    pass


class ColType_UUID(ColType, IKey):
    python_type = ArithUUID


class ColType_Alphanum(ColType, IKey):
    python_type = ArithAlphanumeric


class Native_UUID(ColType_UUID):
    pass


class String_UUID(StringType, ColType_UUID):
    pass


@dataclass
class String_Alphanum(StringType, ColType_Alphanum):
    length: int

    @staticmethod
    def test_value(value: str) -> bool:
        try:
            ArithAlphanumeric(value)
            return True
        except ValueError:
            return False

    def make_value(self, value):
        if len(value) != self.length:
            raise ValueError(f"Expected alphanumeric value of length {self.length}, but got '{value}'.")
        return self.python_type(value, max_len=self.length)


@dataclass
class Text(StringType):
    supported = False


@dataclass
class Integer(NumericType, IKey):
    precision: int = 0
    python_type: type = int

    def __post_init__(self):
        assert self.precision == 0


@dataclass
class UnknownColType(ColType):
    text: str

    supported = False


class AbstractDatabase(ABC):
    @abstractmethod
    def quote(self, s: str):
        "Quote SQL name (implementation specific)"
        ...

    @abstractmethod
    def to_string(self, s: str) -> str:
        "Provide SQL for casting a column to string"
        ...

    @abstractmethod
    def concat(self, l: List[str]) -> str:
        "Provide SQL for concatenating a bunch of column into a string"
        ...

    @abstractmethod
    def timestamp_value(self, t: DbTime) -> str:
        "Provide SQL for the given timestamp value"
        ...

    @abstractmethod
    def md5_to_int(self, s: str) -> str:
        "Provide SQL for computing md5 and returning an int"
        ...

    @abstractmethod
    def offset_limit(self, offset: Optional[int] = None, limit: Optional[int] = None):
        "Provide SQL fragment for limit and offset inside a select"
        ...

    @abstractmethod
    def _query(self, sql_code: str) -> list:
        "Send query to database and return result"
        ...

    @abstractmethod
    def select_table_schema(self, path: DbPath) -> str:
        "Provide SQL for selecting the table schema as (name, type, date_prec, num_prec)"
        ...

    @abstractmethod
    def query_table_schema(self, path: DbPath) -> Dict[str, tuple]:
        """Query the table for its schema for table in 'path', and return {column: tuple}
        where the tuple is (table_name, col_name, type_repr, datetime_precision?, numeric_precision?, numeric_scale?)
        """
        ...

    @abstractmethod
    def _process_table_schema(self, path: DbPath, raw_schema: Dict[str, tuple], filter_columns: Sequence[str]):
        """Process the result of query_table_schema().

        Done in a separate step, to minimize the amount of processed columns.
        Needed because processing each column may:
        * throw errors and warnings
        * query the database to sample values

        """

    @abstractmethod
    def parse_table_name(self, name: str) -> DbPath:
        "Parse the given table name into a DbPath"
        ...

    @abstractmethod
    def close(self):
        "Close connection(s) to the database instance. Querying will stop functioning."
        ...

    @abstractmethod
    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        """Creates an SQL expression, that converts 'value' to a normalized timestamp.

        The returned expression must accept any SQL datetime/timestamp, and return a string.

        Date format: ``YYYY-MM-DD HH:mm:SS.FFFFFF``

        Precision of dates should be rounded up/down according to coltype.rounds
        """
        ...

    @abstractmethod
    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        """Creates an SQL expression, that converts 'value' to a normalized number.

        The returned expression must accept any SQL int/numeric/float, and return a string.

        Floats/Decimals are expected in the format
        "I.P"

        Where I is the integer part of the number (as many digits as necessary),
        and must be at least one digit (0).
        P is the fractional digits, the amount of which is specified with
        coltype.precision. Trailing zeroes may be necessary.
        If P is 0, the dot is omitted.

        Note: We use 'precision' differently than most databases. For decimals,
        it's the same as ``numeric_scale``, and for floats, who use binary precision,
        it can be calculated as ``log10(2**numeric_precision)``.
        """
        ...

    @abstractmethod
    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        """Creates an SQL expression, that converts 'value' to a normalized uuid.

        i.e. just makes sure there is no trailing whitespace.
        """
        ...

    def normalize_value_by_type(self, value: str, coltype: ColType) -> str:
        """Creates an SQL expression, that converts 'value' to a normalized representation.

        The returned expression must accept any SQL value, and return a string.

        The default implementation dispatches to a method according to `coltype`:

        ::

            TemporalType    -> normalize_timestamp()
            FractionalType  -> normalize_number()
            *else*          -> to_string()

            (`Integer` falls in the *else* category)

        """
        if isinstance(coltype, TemporalType):
            return self.normalize_timestamp(value, coltype)
        elif isinstance(coltype, FractionalType):
            return self.normalize_number(value, coltype)
        elif isinstance(coltype, ColType_UUID):
            return self.normalize_uuid(value, coltype)
        return self.to_string(value)

    @abstractmethod
    def _normalize_table_path(self, path: DbPath) -> DbPath:
        ...


Schema = CaseAwareMapping



import logging

from .database_types import *
from .base import ConnectError, Database, import_helper, _query_conn, CHECKSUM_MASK


@import_helper("snowflake")
def import_snowflake():
    import snowflake.connector
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend

    return snowflake, serialization, default_backend


class Snowflake(Database):
    TYPE_CLASSES = {
        # Timestamps
        "TIMESTAMP_NTZ": Timestamp,
        "TIMESTAMP_LTZ": Timestamp,
        "TIMESTAMP_TZ": TimestampTZ,
        # Numbers
        "NUMBER": Decimal,
        "FLOAT": Float,
        # Text
        "TEXT": Text,
    }
    ROUNDS_ON_PREC_LOSS = False

    def __init__(self, *, schema: str, **kw):
        snowflake, serialization, default_backend = import_snowflake()
        logging.getLogger("snowflake.connector").setLevel(logging.WARNING)

        # Ignore the error: snowflake.connector.network.RetryRequest: could not find io module state
        # It's a known issue: https://github.com/snowflakedb/snowflake-connector-python/issues/145
        logging.getLogger("snowflake.connector.network").disabled = True

        assert '"' not in schema, "Schema name should not contain quotes!"
        # If a private key is used, read it from the specified path and pass it as "private_key" to the connector.
        if "key" in kw:
            with open(kw.get("key"), "rb") as key:
                if "password" in kw:
                    raise ConnectError("Cannot use password and key at the same time")
                p_key = serialization.load_pem_private_key(
                    key.read(),
                    password=None,
                    backend=default_backend(),
                )

            kw["private_key"] = p_key.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )

        self._conn = snowflake.connector.connect(schema=f'"{schema}"', **kw)

        self.default_schema = schema

    def close(self):
        self._conn.close()

    def _query(self, sql_code: str) -> list:
        "Uses the standard SQL cursor interface"
        return _query_conn(self._conn, sql_code)

    def quote(self, s: str):
        return f'"{s}"'

    def md5_to_int(self, s: str) -> str:
        return f"BITAND(md5_number_lower64({s}), {CHECKSUM_MASK})"

    def to_string(self, s: str):
        return f"cast({s} as string)"

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)
        return super().select_table_schema((schema, table))

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            timestamp = f"to_timestamp(round(date_part(epoch_nanosecond, {value}::timestamp(9))/1000000000, {coltype.precision}))"
        else:
            timestamp = f"cast({value} as timestamp({coltype.precision}))"

        return f"to_char({timestamp}, 'YYYY-MM-DD HH24:MI:SS.FF6')"

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return self.to_string(f"cast({value} as decimal(38, {coltype.precision}))")



import logging

from .database_types import *
from .base import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, Database, import_helper, _query_conn, parse_table_name


@import_helper("databricks")
def import_databricks():
    import databricks.sql

    return databricks


class Databricks(Database):
    TYPE_CLASSES = {
        # Numbers
        "INT": Integer,
        "SMALLINT": Integer,
        "TINYINT": Integer,
        "BIGINT": Integer,
        "FLOAT": Float,
        "DOUBLE": Float,
        "DECIMAL": Decimal,
        # Timestamps
        "TIMESTAMP": Timestamp,
        # Text
        "STRING": Text,
    }

    ROUNDS_ON_PREC_LOSS = True

    def __init__(
        self,
        http_path: str,
        access_token: str,
        server_hostname: str,
        catalog: str = "hive_metastore",
        schema: str = "default",
        **kwargs,
    ):
        databricks = import_databricks()

        self._conn = databricks.sql.connect(
            server_hostname=server_hostname, http_path=http_path, access_token=access_token
        )

        logging.getLogger("databricks.sql").setLevel(logging.WARNING)

        self.catalog = catalog
        self.default_schema = schema
        self.kwargs = kwargs

    def _query(self, sql_code: str) -> list:
        "Uses the standard SQL cursor interface"
        return _query_conn(self._conn, sql_code)

    def quote(self, s: str):
        return f"`{s}`"

    def md5_to_int(self, s: str) -> str:
        return f"cast(conv(substr(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16, 10) as decimal(38, 0))"

    def to_string(self, s: str) -> str:
        return f"cast({s} as string)"

    def _convert_db_precision_to_digits(self, p: int) -> int:
        # Subtracting 1 due to wierd precision issues
        return max(super()._convert_db_precision_to_digits(p) - 1, 0)

    def query_table_schema(self, path: DbPath) -> Dict[str, tuple]:
        # Databricks has INFORMATION_SCHEMA only for Databricks Runtime, not for Databricks SQL.
        # https://docs.databricks.com/spark/latest/spark-sql/language-manual/information-schema/columns.html
        # So, to obtain information about schema, we should use another approach.

        schema, table = self._normalize_table_path(path)
        with self._conn.cursor() as cursor:
            cursor.columns(catalog_name=self.catalog, schema_name=schema, table_name=table)
            rows = cursor.fetchall()
            if not rows:
                raise RuntimeError(f"{self.name}: Table '{'.'.join(path)}' does not exist, or has no columns")

            d = {r.COLUMN_NAME: r for r in rows}
            assert len(d) == len(rows)
            return d

    def _process_table_schema(self, path: DbPath, raw_schema: Dict[str, tuple], filter_columns: Sequence[str]):
        accept = {i.lower() for i in filter_columns}
        rows = [row for name, row in raw_schema.items() if name.lower() in accept]

        resulted_rows = []
        for row in rows:
            row_type = "DECIMAL" if row.DATA_TYPE == 3 else row.TYPE_NAME
            type_cls = self.TYPE_CLASSES.get(row_type, UnknownColType)

            if issubclass(type_cls, Integer):
                row = (row.COLUMN_NAME, row_type, None, None, 0)

            elif issubclass(type_cls, Float):
                numeric_precision = self._convert_db_precision_to_digits(row.DECIMAL_DIGITS)
                row = (row.COLUMN_NAME, row_type, None, numeric_precision, None)

            elif issubclass(type_cls, Decimal):
                # TYPE_NAME has a format DECIMAL(x,y)
                items = row.TYPE_NAME[8:].rstrip(")").split(",")
                numeric_precision, numeric_scale = int(items[0]), int(items[1])
                row = (row.COLUMN_NAME, row_type, None, numeric_precision, numeric_scale)

            elif issubclass(type_cls, Timestamp):
                row = (row.COLUMN_NAME, row_type, row.DECIMAL_DIGITS, None, None)

            else:
                row = (row.COLUMN_NAME, row_type, None, None, None)

            resulted_rows.append(row)

        col_dict: Dict[str, ColType] = {row[0]: self._parse_type(path, *row) for row in resulted_rows}

        self._refine_coltypes(path, col_dict)
        return col_dict

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        """Databricks timestamp contains no more than 6 digits in precision"""

        if coltype.rounds:
            timestamp = f"cast(round(unix_micros({value}) / 1000000, {coltype.precision}) * 1000000 as bigint)"
            return f"date_format(timestamp_micros({timestamp}), 'yyyy-MM-dd HH:mm:ss.SSSSSS')"

        precision_format = "S" * coltype.precision + "0" * (6 - coltype.precision)
        return f"date_format({value}, 'yyyy-MM-dd HH:mm:ss.{precision_format}')"

    def normalize_number(self, value: str, coltype: NumericType) -> str:
        return self.to_string(f"cast({value} as decimal(38, {coltype.precision}))")

    def parse_table_name(self, name: str) -> DbPath:
        path = parse_table_name(name)
        return self._normalize_table_path(path)

    def close(self):
        self._conn.close()



"""Provides classes for a pseudo-SQL AST that compiles to SQL code
"""

from typing import Sequence, Union, Optional
from datetime import datetime

from runtype import dataclass

from .utils import join_iter, ArithString

from .databases.database_types import AbstractDatabase, DbPath


class Sql:
    pass


SqlOrStr = Union[Sql, str]

CONCAT_SEP = "|"


@dataclass
class Compiler:
    """Provides a set of utility methods for compiling SQL

    For internal use.
    """

    database: AbstractDatabase
    in_select: bool = False  # Compilation

    def quote(self, s: str):
        return self.database.quote(s)

    def compile(self, elem):
        if isinstance(elem, Sql):
            return elem.compile(self)
        elif isinstance(elem, str):
            return elem
        elif isinstance(elem, int):
            return str(elem)
        assert False


@dataclass
class TableName(Sql):
    name: DbPath

    def compile(self, c: Compiler):
        path = c.database._normalize_table_path(self.name)
        return ".".join(map(c.quote, path))


@dataclass
class ColumnName(Sql):
    name: str

    def compile(self, c: Compiler):
        return c.quote(self.name)


@dataclass
class Value(Sql):
    value: object  # Primitive

    def compile(self, c: Compiler):
        if isinstance(self.value, bytes):
            return f"b'{self.value.decode()}'"
        elif isinstance(self.value, str):
            return f"'{self.value}'" % self.value
        elif isinstance(self.value, ArithString):
            return f"'{self.value}'"
        return str(self.value)


@dataclass
class Select(Sql):
    columns: Sequence[SqlOrStr]
    table: SqlOrStr = None
    where: Sequence[SqlOrStr] = None
    order_by: Sequence[SqlOrStr] = None
    group_by: Sequence[SqlOrStr] = None
    limit: int = None

    def compile(self, parent_c: Compiler):
        c = parent_c.replace(in_select=True)
        columns = ", ".join(map(c.compile, self.columns))
        select = f"SELECT {columns}"

        if self.table:
            select += " FROM " + c.compile(self.table)

        if self.where:
            select += " WHERE " + " AND ".join(map(c.compile, self.where))

        if self.group_by:
            select += " GROUP BY " + ", ".join(map(c.compile, self.group_by))

        if self.order_by:
            select += " ORDER BY " + ", ".join(map(c.compile, self.order_by))

        if self.limit is not None:
            select += " " + c.database.offset_limit(0, self.limit)

        if parent_c.in_select:
            select = "(%s)" % select
        return select


@dataclass
class Enum(Sql):
    table: DbPath
    order_by: SqlOrStr

    def compile(self, c: Compiler):
        table = ".".join(map(c.quote, self.table))
        order = c.compile(self.order_by)
        return f"(SELECT *, (row_number() over (ORDER BY {order})) as idx FROM {table} ORDER BY {order}) tmp"


@dataclass
class Checksum(Sql):
    exprs: Sequence[SqlOrStr]

    def compile(self, c: Compiler):
        if len(self.exprs) > 1:
            compiled_exprs = [f"coalesce({c.compile(expr)}, '<null>')" for expr in self.exprs]
            separated = list(join_iter(f"'|'", compiled_exprs))
            expr = c.database.concat(separated)
        else:
            # No need to coalesce - safe to assume that key cannot be null
            (expr,) = self.exprs
            expr = c.compile(expr)
        md5 = c.database.md5_to_int(expr)
        return f"sum({md5})"


@dataclass
class Compare(Sql):
    op: str
    a: SqlOrStr
    b: SqlOrStr

    def compile(self, c: Compiler):
        return f"({c.compile(self.a)} {self.op} {c.compile(self.b)})"


@dataclass
class In(Sql):
    expr: SqlOrStr
    list: Sequence  # List[SqlOrStr]

    def compile(self, c: Compiler):
        elems = ", ".join(map(c.compile, self.list))
        return f"({c.compile(self.expr)} IN ({elems}))"


@dataclass
class Count(Sql):
    column: Optional[SqlOrStr] = None

    def compile(self, c: Compiler):
        if self.column:
            return f"count({c.compile(self.column)})"
        return "count(*)"


@dataclass
class Min(Sql):
    column: SqlOrStr

    def compile(self, c: Compiler):
        return f"min({c.compile(self.column)})"


@dataclass
class Max(Sql):
    column: SqlOrStr

    def compile(self, c: Compiler):
        return f"max({c.compile(self.column)})"


@dataclass
class Time(Sql):
    time: datetime

    def compile(self, c: Compiler):
        return c.database.timestamp_value(self.time)


@dataclass
class Explain(Sql):
    sql: Select

    def compile(self, c: Compiler):
        return f"EXPLAIN {c.compile(self.sql)}"



from ..utils import match_regexps

from .database_types import *
from .base import ThreadedDatabase, import_helper, ConnectError, QueryError
from .base import TIMESTAMP_PRECISION_POS

SESSION_TIME_ZONE = None  # Changed by the tests


@import_helper("oracle")
def import_oracle():
    import cx_Oracle

    return cx_Oracle


class Oracle(ThreadedDatabase):
    TYPE_CLASSES: Dict[str, type] = {
        "NUMBER": Decimal,
        "FLOAT": Float,
        # Text
        "CHAR": Text,
        "NCHAR": Text,
        "NVARCHAR2": Text,
        "VARCHAR2": Text,
    }
    ROUNDS_ON_PREC_LOSS = True

    def __init__(self, *, host, database, thread_count, **kw):
        self.kwargs = dict(dsn=f"{host}/{database}" if database else host, **kw)

        self.default_schema = kw.get("user")

        super().__init__(thread_count=thread_count)

    def create_connection(self):
        self._oracle = import_oracle()
        try:
            c = self._oracle.connect(**self.kwargs)
            if SESSION_TIME_ZONE:
                c.cursor().execute(f"ALTER SESSION SET TIME_ZONE = '{SESSION_TIME_ZONE}'")
            return c
        except Exception as e:
            raise ConnectError(*e.args) from e

    def _query(self, sql_code: str):
        try:
            return super()._query(sql_code)
        except self._oracle.DatabaseError as e:
            raise QueryError(e)

    def md5_to_int(self, s: str) -> str:
        # standard_hash is faster than DBMS_CRYPTO.Hash
        # TODO: Find a way to use UTL_RAW.CAST_TO_BINARY_INTEGER ?
        return f"to_number(substr(standard_hash({s}, 'MD5'), 18), 'xxxxxxxxxxxxxxx')"

    def quote(self, s: str):
        return f"{s}"

    def to_string(self, s: str):
        return f"cast({s} as varchar(1024))"

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)

        return (
            f"SELECT column_name, data_type, 6 as datetime_precision, data_precision as numeric_precision, data_scale as numeric_scale"
            f" FROM ALL_TAB_COLUMNS WHERE table_name = '{table.upper()}' AND owner = '{schema.upper()}'"
        )

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            return f"to_char(cast({value} as timestamp({coltype.precision})), 'YYYY-MM-DD HH24:MI:SS.FF6')"

        if coltype.precision > 0:
            truncated = f"to_char({value}, 'YYYY-MM-DD HH24:MI:SS.FF{coltype.precision}')"
        else:
            truncated = f"to_char({value}, 'YYYY-MM-DD HH24:MI:SS.')"
        return f"RPAD({truncated}, {TIMESTAMP_PRECISION_POS+6}, '0')"

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        # FM999.9990
        format_str = "FM" + "9" * (38 - coltype.precision)
        if coltype.precision:
            format_str += "0." + "9" * (coltype.precision - 1) + "0"
        return f"to_char({value}, '{format_str}')"

    def _parse_type(
        self,
        table_path: DbPath,
        col_name: str,
        type_repr: str,
        datetime_precision: int = None,
        numeric_precision: int = None,
        numeric_scale: int = None,
    ) -> ColType:
        regexps = {
            r"TIMESTAMP\((\d)\) WITH LOCAL TIME ZONE": Timestamp,
            r"TIMESTAMP\((\d)\) WITH TIME ZONE": TimestampTZ,
            r"TIMESTAMP\((\d)\)": Timestamp,
        }

        for m, t_cls in match_regexps(regexps, type_repr):
            precision = int(m.group(1))
            return t_cls(precision=precision, rounds=self.ROUNDS_ON_PREC_LOSS)

        return super()._parse_type(
            table_path, col_name, type_repr, datetime_precision, numeric_precision, numeric_scale
        )

    def offset_limit(self, offset: Optional[int] = None, limit: Optional[int] = None):
        if offset:
            raise NotImplementedError("No support for OFFSET in query")

        return f"FETCH NEXT {limit} ROWS ONLY"

    def concat(self, l: List[str]) -> str:
        joined_exprs = " || ".join(l)
        return f"({joined_exprs})"

    def timestamp_value(self, t: DbTime) -> str:
        return "timestamp '%s'" % t.isoformat(" ")

    def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
        # Cast is necessary for correct MD5 (trimming not enough)
        return f"CAST(TRIM({value}) AS VARCHAR(36))"



from copy import deepcopy
import sys
import time
import json
import logging
from itertools import islice

from data_diff.tracking import disable_tracking

from .utils import remove_password_from_url, safezip, match_like

from .diff_tables import (
    TableSegment,
    TableDiffer,
    DEFAULT_BISECTION_THRESHOLD,
    DEFAULT_BISECTION_FACTOR,
    create_schema,
)
from .databases.connect import connect
from .parse_time import parse_time_before_now, UNITS_STR, ParseError
from .config import apply_config_from_file

import rich
import click

LOG_FORMAT = "[%(asctime)s] %(levelname)s - %(message)s"
DATE_FORMAT = "%H:%M:%S"

COLOR_SCHEME = {
    "+": "green",
    "-": "red",
}


def _remove_passwords_in_dict(d: dict):
    for k, v in d.items():
        if k == "password":
            d[k] = "*" * len(v)
        elif isinstance(v, dict):
            _remove_passwords_in_dict(v)
        elif k.startswith("database"):
            d[k] = remove_password_from_url(v)


def _get_schema(pair):
    db, table_path = pair
    return db.query_table_schema(table_path)


@click.command()
@click.argument("database1", required=False)
@click.argument("table1", required=False)
@click.argument("database2", required=False)
@click.argument("table2", required=False)
@click.option("-k", "--key-column", default=None, help="Name of primary key column. Default='id'.")
@click.option("-t", "--update-column", default=None, help="Name of updated_at/last_updated column")
@click.option(
    "-c",
    "--columns",
    default=[],
    multiple=True,
    help="Names of extra columns to compare."
    "Can be used more than once in the same command. "
    "Accepts a name or a pattern like in SQL. Example: -c col% -c another_col",
)
@click.option("-l", "--limit", default=None, help="Maximum number of differences to find")
@click.option("--bisection-factor", default=None, help=f"Segments per iteration. Default={DEFAULT_BISECTION_FACTOR}.")
@click.option(
    "--bisection-threshold",
    default=None,
    help=f"Minimal bisection threshold. Below it, data-diff will download the data and compare it locally. Default={DEFAULT_BISECTION_THRESHOLD}.",
)
@click.option(
    "--min-age",
    default=None,
    help="Considers only rows older than specified. Useful for specifying replication lag."
    "Example: --min-age=5min ignores rows from the last 5 minutes. "
    f"\nValid units: {UNITS_STR}",
)
@click.option("--max-age", default=None, help="Considers only rows younger than specified. See --min-age.")
@click.option("-s", "--stats", is_flag=True, help="Print stats instead of a detailed diff")
@click.option("-d", "--debug", is_flag=True, help="Print debug info")
@click.option("--json", "json_output", is_flag=True, help="Print JSONL output for machine readability")
@click.option("-v", "--verbose", is_flag=True, help="Print extra info")
@click.option("-i", "--interactive", is_flag=True, help="Confirm queries, implies --debug")
@click.option("--no-tracking", is_flag=True, help="data-diff sends home anonymous usage data. Use this to disable it.")
@click.option(
    "--case-sensitive",
    is_flag=True,
    help="Column names are treated as case-sensitive. Otherwise, data-diff corrects their case according to schema.",
)
@click.option(
    "-j",
    "--threads",
    default=None,
    help="Number of worker threads to use per database. Default=1. "
    "A higher number will increase performance, but take more capacity from your database. "
    "'serial' guarantees a single-threaded execution of the algorithm (useful for debugging).",
)
@click.option("-w", "--where", default=None, help="An additional 'where' expression to restrict the search space.")
@click.option(
    "--conf",
    default=None,
    help="Path to a configuration.toml file, to provide a default configuration, and a list of possible runs.",
)
@click.option(
    "--run",
    default=None,
    help="Name of run-configuration to run. If used, CLI arguments for database and table must be omitted.",
)
def main(conf, run, **kw):
    if conf:
        kw = apply_config_from_file(conf, run, kw)
    return _main(**kw)


def _main(
    database1,
    table1,
    database2,
    table2,
    key_column,
    update_column,
    columns,
    limit,
    bisection_factor,
    bisection_threshold,
    min_age,
    max_age,
    stats,
    debug,
    verbose,
    interactive,
    no_tracking,
    threads,
    case_sensitive,
    json_output,
    where,
    threads1=None,
    threads2=None,
    __conf__=None,
):

    if no_tracking:
        disable_tracking()

    if interactive:
        debug = True

    if debug:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT, datefmt=DATE_FORMAT)
        if __conf__:
            __conf__ = deepcopy(__conf__)
            _remove_passwords_in_dict(__conf__)
            logging.debug(f"Applied run configuration: {__conf__}")
    elif verbose:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)

    if limit and stats:
        logging.error("Cannot specify a limit when using the -s/--stats switch")
        return

    key_column = key_column or "id"
    bisection_factor = DEFAULT_BISECTION_FACTOR if bisection_factor is None else int(bisection_factor)
    bisection_threshold = DEFAULT_BISECTION_THRESHOLD if bisection_threshold is None else int(bisection_threshold)

    threaded = True
    if threads is None:
        threads = 1
    elif isinstance(threads, str) and threads.lower() == "serial":
        assert not (threads1 or threads2)
        threaded = False
        threads = 1
    else:
        try:
            threads = int(threads)
        except ValueError:
            logging.error("Error: threads must be a number, or 'serial'.")
            return
        if threads < 1:
            logging.error("Error: threads must be >= 1")
            return

    db1 = connect(database1, threads1 or threads)
    db2 = connect(database2, threads2 or threads)
    dbs = db1, db2

    if interactive:
        for db in dbs:
            db.enable_interactive()

    start = time.monotonic()

    try:
        options = dict(
            min_update=max_age and parse_time_before_now(max_age),
            max_update=min_age and parse_time_before_now(min_age),
            case_sensitive=case_sensitive,
            where=where,
        )
    except ParseError as e:
        logging.error("Error while parsing age expression: %s" % e)
        return

    differ = TableDiffer(
        bisection_factor=bisection_factor,
        bisection_threshold=bisection_threshold,
        threaded=threaded,
        max_threadpool_size=threads and threads * 2,
        debug=debug,
    )

    table_names = table1, table2
    table_paths = [db.parse_table_name(t) for db, t in safezip(dbs, table_names)]

    schemas = list(differ._thread_map(_get_schema, safezip(dbs, table_paths)))
    schema1, schema2 = schemas = [
        create_schema(db, table_path, schema, case_sensitive)
        for db, table_path, schema in safezip(dbs, table_paths, schemas)
    ]

    mutual = schema1.keys() & schema2.keys()  # Case-aware, according to case_sensitive
    logging.debug(f"Available mutual columns: {mutual}")

    expanded_columns = set()
    for c in columns:
        match = set(match_like(c, mutual))
        if not match:
            m1 = None if any(match_like(c, schema1.keys())) else f"{db1}/{table1}"
            m2 = None if any(match_like(c, schema2.keys())) else f"{db2}/{table2}"
            not_matched = ", ".join(m for m in [m1, m2] if m)
            raise ValueError(f"Column {c} not found in: {not_matched}")

        expanded_columns |= match

    columns = tuple(expanded_columns - {key_column, update_column})

    logging.info(f"Diffing columns: key={key_column} update={update_column} extra={columns}")

    segments = [
        TableSegment(db, table_path, key_column, update_column, columns, **options)._with_raw_schema(raw_schema)
        for db, table_path, raw_schema in safezip(dbs, table_paths, schemas)
    ]

    diff_iter = differ.diff_tables(*segments)

    if limit:
        diff_iter = islice(diff_iter, int(limit))

    if stats:
        diff = list(diff_iter)
        unique_diff_count = len({i[0] for _, i in diff})
        max_table_count = max(differ.stats["table1_count"], differ.stats["table2_count"])
        percent = 100 * unique_diff_count / (max_table_count or 1)
        plus = len([1 for op, _ in diff if op == "+"])
        minus = len([1 for op, _ in diff if op == "-"])

        if json_output:
            json_output = {
                "different_rows": len(diff),
                "different_percent": percent,
                "different_+": plus,
                "different_-": minus,
                "total": max_table_count,
            }
            print(json.dumps(json_output))
        else:
            print(f"Diff-Total: {len(diff)} changed rows out of {max_table_count}")
            print(f"Diff-Percent: {percent:.14f}%")
            print(f"Diff-Split: +{plus}  -{minus}")
    else:
        for op, values in diff_iter:
            color = COLOR_SCHEME[op]

            if json_output:
                jsonl = json.dumps([op, list(values)])
                rich.print(f"[{color}]{jsonl}[/{color}]")
            else:
                text = f"{op} {', '.join(values)}"
                rich.print(f"[{color}]{text}[/{color}]")

            sys.stdout.flush()

    end = time.monotonic()

    logging.info(f"Duration: {end-start:.2f} seconds.")


if __name__ == "__main__":
    main()



from .database_types import *
from .postgresql import PostgreSQL, MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, TIMESTAMP_PRECISION_POS


class Redshift(PostgreSQL):
    TYPE_CLASSES = {
        **PostgreSQL.TYPE_CLASSES,
        "double": Float,
        "real": Float,
    }

    def md5_to_int(self, s: str) -> str:
        return f"strtol(substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16)::decimal(38)"

    def normalize_timestamp(self, value: str, coltype: TemporalType) -> str:
        if coltype.rounds:
            timestamp = f"{value}::timestamp(6)"
            # Get seconds since epoch. Redshift doesn't support milli- or micro-seconds.
            secs = f"timestamp 'epoch' + round(extract(epoch from {timestamp})::decimal(38)"
            # Get the milliseconds from timestamp.
            ms = f"extract(ms from {timestamp})"
            # Get the microseconds from timestamp, without the milliseconds!
            us = f"extract(us from {timestamp})"
            # epoch = Total time since epoch in microseconds.
            epoch = f"{secs}*1000000 + {ms}*1000 + {us}"
            timestamp6 = (
                f"to_char({epoch}, -6+{coltype.precision}) * interval '0.000001 seconds', 'YYYY-mm-dd HH24:MI:SS.US')"
            )
        else:
            timestamp6 = f"to_char({value}::timestamp(6), 'YYYY-mm-dd HH24:MI:SS.US')"
        return (
            f"RPAD(LEFT({timestamp6}, {TIMESTAMP_PRECISION_POS+coltype.precision}), {TIMESTAMP_PRECISION_POS+6}, '0')"
        )

    def normalize_number(self, value: str, coltype: FractionalType) -> str:
        return self.to_string(f"{value}::decimal(38,{coltype.precision})")

    def concat(self, l: List[str]) -> str:
        joined_exprs = " || ".join(l)
        return f"({joined_exprs})"

    def select_table_schema(self, path: DbPath) -> str:
        schema, table = self._normalize_table_path(path)

        return (
            "SELECT column_name, data_type, datetime_precision, numeric_precision, numeric_scale FROM information_schema.columns "
            f"WHERE table_name = '{table.lower()}' AND table_schema = '{schema.lower()}'"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['data_diff', 'data_diff.databases']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1,<9.0',
 'dsnparse',
 'rich',
 'runtype>=0.2.6,<0.3.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['data-diff = data_diff.__main__:main']}

setup_kwargs = {
    'name': 'data-diff',
    'version': '0.2.5',
    'description': 'Command-line tool and Python library to efficiently diff rows across two different databases.',
    'long_description': '# **data-diff**\n\n**data-diff is in shape to be run in production, but also under development. If\nyou run into issues or bugs, please [open an issue](https://github.com/datafold/data-diff/issues/new/choose) and we\'ll help you out ASAP! You can\nalso find us in `#tools-data-diff` in the [Locally Optimistic Slack][slack], and we\'d love to hear about your experience using data-diff.**\n\n**data-diff** is a command-line tool and Python library to efficiently diff\nrows across two different databases.\n\n* ⇄  Verifies across [many different databases][dbs] (e.g. PostgreSQL -> Snowflake)\n* 🔍 Outputs [diff of rows](#example-command-and-output) in detail\n* 🚨 Simple CLI/API to create monitoring and alerts\n* 🔁 Bridges column types of different formats and levels of precision (e.g. Double ⇆ Float ⇆ Decimal)\n* 🔥 Verify 25M+ rows in <10s, and 1B+ rows in ~5min.\n* ♾️  Works for tables with 10s of billions of rows\n\n**data-diff** splits the table into smaller segments, then checksums each\nsegment in both databases. When the checksums for a segment aren\'t equal, it\nwill further divide that segment into yet smaller segments, checksumming those\nuntil it gets to the differing row(s). See [Technical Explanation][tech-explain] for more\ndetails.\n\nThis approach has performance within an order of magnitude of `count(*)` when\nthere are few/no changes, but is able to output each differing row! By pushing\nthe compute into the databases, it\'s _much_ faster than querying for and\ncomparing every row.\n\n![Performance for 100M rows](https://user-images.githubusercontent.com/97400/175182987-a3900d4e-c097-4732-a4e9-19a40fac8cdc.png)\n\n**†:** The implementation for downloading all rows that `data-diff` and\n`count(*)` is compared to is not optimal. It is a single Python multi-threaded\nprocess. The performance is fairly driver-specific, e.g. PostgreSQL\'s performs 10x\nbetter than MySQL.\n\n## Table of Contents\n\n- [**data-diff**](#data-diff)\n  - [Table of Contents](#table-of-contents)\n  - [Common use-cases](#common-use-cases)\n  - [Example Command and Output](#example-command-and-output)\n  - [Supported Databases](#supported-databases)\n- [How to install](#how-to-install)\n  - [Install drivers](#install-drivers)\n- [How to use](#how-to-use)\n  - [How to use from the command-line](#how-to-use-from-the-command-line)\n  - [How to use from Python](#how-to-use-from-python)\n- [Technical Explanation](#technical-explanation)\n  - [Performance Considerations](#performance-considerations)\n- [Anonymous Tracking](#anonymous-tracking)\n- [Development Setup](#development-setup)\n- [License](#license)\n\n## Common use-cases\n\n* **Verify data migrations.** Verify that all data was copied when doing a\n  critical data migration. For example, migrating from Heroku PostgreSQL to Amazon RDS.\n* **Verifying data pipelines.** Moving data from a relational database to a\n  warehouse/data lake with Fivetran, Airbyte, Debezium, or some other pipeline.\n* **Alerting and maintaining data integrity SLOs.** You can create and monitor\n  your SLO of e.g. 99.999% data integrity, and alert your team when data is\n  missing.\n* **Debugging complex data pipelines.** When data gets lost in pipelines that\n  may span a half-dozen systems, without verifying each intermediate datastore\n  it\'s extremely difficult to track down where a row got lost.\n* **Detecting hard deletes for an `updated_at`-based pipeline**. If you\'re\n  copying data to your warehouse based on an `updated_at`-style column, then\n  you\'ll miss hard-deletes that **data-diff** can find for you.\n* **Make your replication self-healing.** You can use **data-diff** to\n  self-heal by using the diff output to write/update rows in the target\n  database.\n\n## Example Command and Output\n\nBelow we run a comparison with the CLI for 25M rows in PostgreSQL where the\nright-hand table is missing single row with `id=12500048`:\n\n```\n$ data-diff \\\n    postgresql://user:password@localhost/database rating \\\n    postgresql://user:password@localhost/database rating_del1 \\\n    --bisection-threshold 100000 \\ # for readability, try default first\n    --bisection-factor 6 \\ # for readability, try default first\n    --update-column timestamp \\\n    --verbose\n\n    # Consider running with --interactive the first time.\n    # Runs `EXPLAIN` for you to verify the queries are using indexes.\n    # --interactive\n[10:15:00] INFO - Diffing tables | segments: 6, bisection threshold: 100000.\n[10:15:00] INFO - . Diffing segment 1/6, key-range: 1..4166683, size: 4166682\n[10:15:03] INFO - . Diffing segment 2/6, key-range: 4166683..8333365, size: 4166682\n[10:15:06] INFO - . Diffing segment 3/6, key-range: 8333365..12500047, size: 4166682\n[10:15:09] INFO - . Diffing segment 4/6, key-range: 12500047..16666729, size: 4166682\n[10:15:12] INFO - . . Diffing segment 1/6, key-range: 12500047..13194494, size: 694447\n[10:15:13] INFO - . . . Diffing segment 1/6, key-range: 12500047..12615788, size: 115741\n[10:15:13] INFO - . . . . Diffing segment 1/6, key-range: 12500047..12519337, size: 19290\n[10:15:13] INFO - . . . . Diff found 1 different rows.\n[10:15:13] INFO - . . . . Diffing segment 2/6, key-range: 12519337..12538627, size: 19290\n[10:15:13] INFO - . . . . Diffing segment 3/6, key-range: 12538627..12557917, size: 19290\n[10:15:13] INFO - . . . . Diffing segment 4/6, key-range: 12557917..12577207, size: 19290\n[10:15:13] INFO - . . . . Diffing segment 5/6, key-range: 12577207..12596497, size: 19290\n[10:15:13] INFO - . . . . Diffing segment 6/6, key-range: 12596497..12615788, size: 19291\n[10:15:13] INFO - . . . Diffing segment 2/6, key-range: 12615788..12731529, size: 115741\n[10:15:13] INFO - . . . Diffing segment 3/6, key-range: 12731529..12847270, size: 115741\n[10:15:13] INFO - . . . Diffing segment 4/6, key-range: 12847270..12963011, size: 115741\n[10:15:14] INFO - . . . Diffing segment 5/6, key-range: 12963011..13078752, size: 115741\n[10:15:14] INFO - . . . Diffing segment 6/6, key-range: 13078752..13194494, size: 115742\n[10:15:14] INFO - . . Diffing segment 2/6, key-range: 13194494..13888941, size: 694447\n[10:15:14] INFO - . . Diffing segment 3/6, key-range: 13888941..14583388, size: 694447\n[10:15:15] INFO - . . Diffing segment 4/6, key-range: 14583388..15277835, size: 694447\n[10:15:15] INFO - . . Diffing segment 5/6, key-range: 15277835..15972282, size: 694447\n[10:15:15] INFO - . . Diffing segment 6/6, key-range: 15972282..16666729, size: 694447\n+ (12500048, 1268104625)\n[10:15:16] INFO - . Diffing segment 5/6, key-range: 16666729..20833411, size: 4166682\n[10:15:19] INFO - . Diffing segment 6/6, key-range: 20833411..25000096, size: 4166685\n```\n\n## Supported Databases\n\n| Database      | Connection string                                                                                                                   | Status |\n|---------------|-------------------------------------------------------------------------------------------------------------------------------------|--------|\n| PostgreSQL >=10    | `postgresql://<user>:<password>@<host>:5432/<database>`                                                                         |  💚    |\n| MySQL         | `mysql://<user>:<password>@<hostname>:5432/<database>`                                                                              |  💚    |\n| Snowflake     | `"snowflake://<user>[:<password>]@<account>/<database>/<SCHEMA>?warehouse=<WAREHOUSE>&role=<role>[&authenticator=externalbrowser]"` |  💚    |\n| Oracle        | `oracle://<username>:<password>@<hostname>/database`                                                                                |  💛    |\n| BigQuery      | `bigquery://<project>/<dataset>`                                                                                                    |  💛    |\n| Redshift      | `redshift://<username>:<password>@<hostname>:5439/<database>`                                                                       |  💛    |\n| Presto        | `presto://<username>:<password>@<hostname>:8080/<database>`                                                                         |  💛    |\n| Databricks    | `databricks://<http_path>:<access_token>@<server_hostname>/<catalog>/<schema>`                                                      |  💛    |\n| Trino         | `trino://<username>:<password>@<hostname>:8080/<database>`                                                                          |  💛    |\n| ElasticSearch |                                                                                                                                     |  📝    |\n| Planetscale   |                                                                                                                                     |  📝    |\n| Clickhouse    |                                                                                                                                     |  📝    |\n| Pinot         |                                                                                                                                     |  📝    |\n| Druid         |                                                                                                                                     |  📝    |\n| Kafka         |                                                                                                                                     |  📝    |\n\n* 💚: Implemented and thoroughly tested.\n* 💛: Implemented, but not thoroughly tested yet.\n* ⏳: Implementation in progress.\n* 📝: Implementation planned. Contributions welcome.\n\nIf a database is not on the list, we\'d still love to support it. Open an issue\nto discuss it.\n\nNote: Because URLs allow many special characters, and may collide with the syntax of your command-line,\nit\'s recommended to surround them with quotes. Alternatively, you may provide them in a TOML file via the `--config` option.\n\n\n# How to install\n\nRequires Python 3.7+ with pip.\n\n```pip install data-diff```\n\n## Install drivers\n\nTo connect to a database, we need to have its driver installed, in the form of a Python library.\n\nWhile you may install them manually, we offer an easy way to install them along with data-diff<sup>*</sup>:\n\n- `pip install \'data-diff[mysql]\'`\n\n- `pip install \'data-diff[postgresql]\'`\n\n- `pip install \'data-diff[snowflake]\'`\n\n- `pip install \'data-diff[presto]\'`\n\n- `pip install \'data-diff[oracle]\'`\n\n- `pip install \'data-diff[trino]\'`\n\n- For BigQuery, see: https://pypi.org/project/google-cloud-bigquery/\n\n\nUsers can also install several drivers at once:\n\n```pip install \'data-diff[mysql,postgresql,snowflake]\'```\n\n_<sup>*</sup> Some drivers have dependencies that cannot be installed using `pip` and still need to be installed manually._\n\n\n### Install Psycopg2\n\nIn order to run Postgresql, you\'ll need `psycopg2`. This Python package requires some additional dependencies described in their [documentation](https://www.psycopg.org/docs/install.html#build-prerequisites).\nAn easy solution is to install [psycopg2-binary](https://www.psycopg.org/docs/install.html#quick-install) by running:\n\n```pip install psycopg2-binary```\n\nWhich comes with a pre-compiled binary and does not require additonal prerequisites. However, note that for production use it is adviced to use `psycopg2`.\n\n\n# How to use\n\n## How to use from the command-line\n\nUsage: `data-diff DB1_URI TABLE1_NAME DB2_URI TABLE2_NAME [OPTIONS]`\n\nSee the [example command](#example-command-and-output) and the [sample\nconnection strings](#supported-databases).\n\nNote that for some databases, the arguments that you enter in the command line\nmay be case-sensitive. This is the case for the Snowflake schema and table names.\n\nOptions:\n\n  - `--help` - Show help message and exit.\n  - `-k` or `--key-column` - Name of the primary key column\n  - `-t` or `--update-column` - Name of updated_at/last_updated column\n  - `-c` or `--columns` - Names of extra columns to compare.  Can be used more than once in the same command.\n                          Accepts a name or a pattern like in SQL.\n                          Example: `-c col% -c another_col -c %foorb.r%`\n  - `-l` or `--limit` - Maximum number of differences to find (limits maximum bandwidth and runtime)\n  - `-s` or `--stats` - Print stats instead of a detailed diff\n  - `-d` or `--debug` - Print debug info\n  - `-v` or `--verbose` - Print extra info\n  - `-i` or `--interactive` - Confirm queries, implies `--debug`\n  - `--json` - Print JSONL output for machine readability\n  - `--min-age` - Considers only rows older than specified. Useful for specifying replication lag.\n                  Example: `--min-age=5min` ignores rows from the last 5 minutes.\n                  Valid units: `d, days, h, hours, min, minutes, mon, months, s, seconds, w, weeks, y, years`\n  - `--max-age` - Considers only rows younger than specified. See `--min-age`.\n  - `--bisection-factor` - Segments per iteration. When set to 2, it performs binary search.\n  - `--bisection-threshold` - Minimal bisection threshold. i.e. maximum size of pages to diff locally.\n  - `-j` or `--threads` - Number of worker threads to use per database. Default=1.\n  - `-w`, `--where` - An additional \'where\' expression to restrict the search space.\n  - `--conf`, `--run` - Specify the run and configuration from a TOML file. (see below)\n  - `--no-tracking` - data-diff sends home anonymous usage data. Use this to disable it.\n\n\n### How to use with a configuration file\n\nData-diff lets you load the configuration for a run from a TOML file.\n\nReasons to use a configuration file:\n\n- Convenience - Set-up the parameters for diffs that need to run often\n\n- Easier and more readable - you can define the database connection settings as config values, instead of in a URI.\n\n- Gives you fine-grained control over the settings switches, without requiring any Python code.\n\nUse `--conf` to specify that path to the configuration file. data-diff will load the settings from `run.default`, if it\'s defined.\n\nThen you can, optionally, use `--run` to choose to load the settings of a specific run, and override the settings `run.default`. (all runs extend `run.default`, like inheritance).\n\nFinally, CLI switches have the final say, and will override the settings defined by the configuration file, and the current run.\n\nExample TOML file:\n\n```toml\n# Specify the connection params to the test database.\n[database.test_postgresql]\ndriver = "postgresql"\nuser = "postgres"\npassword = "Password1"\n\n# Specify the default run params\n[run.default]\nupdate_column = "timestamp"\nverbose = true\n\n# Specify params for a run \'test_diff\'.\n[run.test_diff]\nverbose = false\n# Source 1 ("left")\n1.database = "test_postgresql"                      # Use options from database.test_postgresql\n1.table = "rating"\n# Source 2 ("right")\n2.database = "postgresql://postgres:Password1@/"    # Use URI like in the CLI\n2.table = "rating_del1"\n```\n\nIn this example, running `data-diff --conf myconfig.toml --run test_diff` will compare between `rating` and `rating_del1`.\nIt will use the `timestamp` column as the update column, as specified in `run.default`. However, it won\'t be verbose, since that\nflag is overwritten to `false`.\n\nRunning it with `data-diff --conf myconfig.toml --run test_diff -v` will set verbose back to `true`.\n\n\n## How to use from Python\n\nAPI reference: [https://data-diff.readthedocs.io/en/latest/](https://data-diff.readthedocs.io/en/latest/)\n\nExample:\n\n```python\n# Optional: Set logging to display the progress of the diff\nimport logging\nlogging.basicConfig(level=logging.INFO)\n\nfrom data_diff import connect_to_table, diff_tables\n\ntable1 = connect_to_table("postgresql:///", "table_name", "id")\ntable2 = connect_to_table("mysql:///", "table_name", "id")\n\nfor different_row in diff_tables(table1, table2):\n    plus_or_minus, columns = different_row\n    print(plus_or_minus, columns)\n```\n\nRun `help(diff_tables)` or [read the docs](https://data-diff.readthedocs.io/en/latest/) to learn about the different options.\n\n# Technical Explanation\n\nIn this section we\'ll be doing a walk-through of exactly how **data-diff**\nworks, and how to tune `--bisection-factor` and `--bisection-threshold`.\n\nLet\'s consider a scenario with an `orders` table with 1M rows. Fivetran is\nreplicating it contionously from PostgreSQL to Snowflake:\n\n```\n┌─────────────┐                        ┌─────────────┐\n│ PostgreSQL  │                        │  Snowflake  │\n├─────────────┤                        ├─────────────┤\n│             │                        │             │\n│             │                        │             │\n│             │  ┌─────────────┐       │ table with  │\n│ table with  ├──┤ replication ├──────▶│ ?maybe? all │\n│lots of rows!│  └─────────────┘       │  the same   │\n│             │                        │    rows.    │\n│             │                        │             │\n│             │                        │             │\n│             │                        │             │\n└─────────────┘                        └─────────────┘\n```\n\nIn order to check whether the two tables are the same, **data-diff** splits\nthe table into `--bisection-factor=10` segments.\n\nWe also have to choose which columns we want to checksum. In our case, we care\nabout the primary key, `--key-column=id` and the update column\n`--update-column=updated_at`. `updated_at` is updated every time the row is, and\nwe have an index on it.\n\n**data-diff** starts by querying both databases for the `min(id)` and `max(id)`\nof the table. Then it splits the table into `--bisection-factor=10` segments of\n`1M/10 = 100K` keys each:\n\n```\n┌──────────────────────┐              ┌──────────────────────┐\n│     PostgreSQL       │              │      Snowflake       │\n├──────────────────────┤              ├──────────────────────┤\n│      id=1..100k      │              │      id=1..100k      │\n├──────────────────────┤              ├──────────────────────┤\n│    id=100k..200k     │              │    id=100k..200k     │\n├──────────────────────┤              ├──────────────────────┤\n│    id=200k..300k     ├─────────────▶│    id=200k..300k     │\n├──────────────────────┤              ├──────────────────────┤\n│    id=300k..400k     │              │    id=300k..400k     │\n├──────────────────────┤              ├──────────────────────┤\n│         ...          │              │         ...          │\n├──────────────────────┤              ├──────────────────────┤\n│      900k..100k      │              │      900k..100k      │\n└───────────────────▲──┘              └▲─────────────────────┘\n                    ┃                  ┃\n                    ┃                  ┃\n                    ┃ checksum queries ┃\n                    ┃                  ┃\n                  ┌─┻──────────────────┻────┐\n                  │        data-diff        │\n                  └─────────────────────────┘\n```\n\nNow **data-diff** will start running `--threads=1` queries in parallel that\nchecksum each segment. The queries for checksumming each segment will look\nsomething like this, depending on the database:\n\n```sql\nSELECT count(*),\n    sum(cast(conv(substring(md5(concat(cast(id as char), cast(timestamp as char))), 18), 16, 10) as unsigned))\nFROM `rating_del1`\nWHERE (id >= 1) AND (id < 100000)\n```\n\nThis keeps the amount of data that has to be transferred between the databases\nto a minimum, making it very performant! Additionally, if you have an index on\n`updated_at` (highly recommended) then the query will be fast as the database\nonly has to do a partial index scan between `id=1..100k`.\n\nIf you are not sure whether the queries are using an index, you can run it with\n`--interactive`. This puts **data-diff** in interactive mode where it shows an\n`EXPLAIN` before executing each query, requiring confirmation to proceed.\n\nAfter running the checksum queries on both sides, we see that all segments\nare the same except `id=100k..200k`:\n\n```\n┌──────────────────────┐              ┌──────────────────────┐\n│     PostgreSQL       │              │      Snowflake       │\n├──────────────────────┤              ├──────────────────────┤\n│    checksum=0102     │              │    checksum=0102     │\n├──────────────────────┤   mismatch!  ├──────────────────────┤\n│    checksum=ffff     ◀──────────────▶    checksum=aaab     │\n├──────────────────────┤              ├──────────────────────┤\n│    checksum=abab     │              │    checksum=abab     │\n├──────────────────────┤              ├──────────────────────┤\n│    checksum=f0f0     │              │    checksum=f0f0     │\n├──────────────────────┤              ├──────────────────────┤\n│         ...          │              │         ...          │\n├──────────────────────┤              ├──────────────────────┤\n│    checksum=9494     │              │    checksum=9494     │\n└──────────────────────┘              └──────────────────────┘\n```\n\nNow **data-diff** will do exactly as it just did for the _whole table_ for only\nthis segment: Split it into `--bisection-factor` segments.\n\nHowever, this time, because each segment has `100k/10=10k` entries, which is\nless than the `--bisection-threshold` it will pull down every row in the segment\nand compare them in memory in **data-diff**.\n\n```\n┌──────────────────────┐              ┌──────────────────────┐\n│     PostgreSQL       │              │      Snowflake       │\n├──────────────────────┤              ├──────────────────────┤\n│    id=100k..110k     │              │    id=100k..110k     │\n├──────────────────────┤              ├──────────────────────┤\n│    id=110k..120k     │              │    id=110k..120k     │\n├──────────────────────┤              ├──────────────────────┤\n│    id=120k..130k     │              │    id=120k..130k     │\n├──────────────────────┤              ├──────────────────────┤\n│    id=130k..140k     │              │    id=130k..140k     │\n├──────────────────────┤              ├──────────────────────┤\n│         ...          │              │         ...          │\n├──────────────────────┤              ├──────────────────────┤\n│      190k..200k      │              │      190k..200k      │\n└──────────────────────┘              └──────────────────────┘\n```\n\nFinally **data-diff** will output the `(id, updated_at)` for each row that was different:\n\n```\n(122001, 1653672821)\n```\n\nIf you pass `--stats` you\'ll see e.g. what % of rows were different.\n\n## Performance Considerations\n\n* Ensure that you have indexes on the columns you are comparing. Preferably a\n  compound index. You can run with `--interactive` to see an `EXPLAIN` for the\n  queries.\n* Consider increasing the number of simultaneous threads executing\n  queries per database with `--threads`. For databases that limit concurrency\n  per query, e.g. PostgreSQL/MySQL, this can improve performance dramatically.\n* If you are only interested in _whether_ something changed, pass `--limit 1`.\n  This can be useful if changes are very rare. This is often faster than doing a\n  `count(*)`, for the reason mentioned above.\n* If the table is _very_ large, consider a larger `--bisection-factor`. Explained in\n  the [technical explanation][tech-explain]. Otherwise you may run into timeouts.\n* If there are a lot of changes, consider a larger `--bisection-threshold`.\n  Explained in the [technical explanation][tech-explain].\n* If there are very large gaps in your key column, e.g. 10s of millions of\n  continuous rows missing, then **data-diff** may perform poorly doing lots of\n  queries for ranges of rows that do not exist (see [technical\n  explanation][tech-explain]). We have ideas on how to tackle this issue, which we have\n  yet to implement. If you\'re experiencing this effect, please open an issue and we\n  will prioritize it.\n* The fewer columns you verify (passed with `--columns`), the faster\n  **data-diff** will be. On one extreme you can verify every column, on the\n  other you can verify _only_ `updated_at`, if you trust it enough. You can also\n  _only_ verify `id` if you\'re interested in only presence, e.g. to detect\n  missing hard deletes. You can do also do a hybrid where you verify\n  `updated_at` and the most critical value, e.g a money value in `amount` but\n  not verify a large serialized column like `json_settings`.\n* We have ideas for making **data-diff** even faster that\n  we haven\'t implemented yet: faster checksums by reducing type-casts\n  and using a faster hash than MD5, dynamic adaptation of\n  `bisection_factor`/`threads`/`bisection_threshold` (especially with large key\n  gaps), and improvements to bypass Python/driver performance limitations when\n  comparing huge amounts of rows locally (i.e. for very high `bisection_threshold` values).\n\n# Usage Analytics\n\ndata-diff collects anonymous usage data to help our team improve the tool and to apply development efforts to where our users need them most.\n\nWe capture two events, one when the data-diff run starts and one when it is finished. No user data or potentially sensitive information is or ever will be collected. The captured data is limited to:\n\n- Operating System and Python version\n\n- Types of databases used (postgresql, mysql, etc.)\n\n- Sizes of tables diffed, run time, and diff row count (numbers only)\n\n- Error message, if any, truncated to the first 20 characters.\n\n- A persistent UUID to indentify the session, stored in `~/.datadiff.toml`\n\nIf you do not wish to participate, the tracking can be easily disabled with one of the following methods:\n\n* In the CLI, use the `--no-tracking` flag.\n\n* In the config file, set `no_tracking = true` (for example, under `[run.default]`)\n\n* If you\'re using the Python API:\n\n```python\nimport data_diff\ndata_diff.disable_tracking()    # Call this first, before making any API calls\n\n# Connect and diff your tables without any tracking\n```\n\n\n# Development Setup\n\nThe development setup centers around using `docker-compose` to boot up various\ndatabases, and then inserting data into them.\n\nFor Mac for performance of Docker, we suggest enabling in the UI:\n\n* Use new Virtualization Framework\n* Enable VirtioFS accelerated directory sharing\n\n**1. Install Data Diff**\n\nWhen developing/debugging, it\'s recommended to install dependencies and run it\ndirectly with `poetry` rather than go through the package.\n\n```\n$ brew install mysql postgresql # MacOS dependencies for C bindings\n$ apt-get install libpq-dev libmysqlclient-dev # Debian dependencies\n\n$ pip install poetry # Python dependency isolation tool\n$ poetry install # Install dependencies\n```\n**2. Start Databases**\n\n[Install **docker-compose**][docker-compose] if you haven\'t already.\n\n```shell-session\n$ docker-compose up -d mysql postgres # run mysql and postgres dbs in background\n```\n\n[docker-compose]: https://docs.docker.com/compose/install/\n\n**3. Run Unit Tests**\n\nThere are more than 1000 tests for all the different type and database\ncombinations, so we recommend using `unittest-parallel` that\'s installed as a\ndevelopment dependency.\n\n```shell-session\n$ poetry run unittest-parallel -j 16 #  run all tests\n$ poetry run python -m unittest -k <test> #  run individual test\n```\n\n**4. Seed the Database(s) (optional)**\n\nFirst, download the CSVs of seeding data:\n\n```shell-session\n$ curl https://datafold-public.s3.us-west-2.amazonaws.com/1m.csv -o dev/ratings.csv\n\n# For a larger data-set (but takes 25x longer to import):\n# - curl https://datafold-public.s3.us-west-2.amazonaws.com/25m.csv -o dev/ratings.csv\n```\n\nNow you can insert it into the testing database(s):\n\n```shell-session\n# It\'s optional to seed more than one to run data-diff(1) against.\n$ poetry run preql -f dev/prepare_db.pql mysql://mysql:Password1@127.0.0.1:3306/mysql\n$ poetry run preql -f dev/prepare_db.pql postgresql://postgres:Password1@127.0.0.1:5432/postgres\n\n# Cloud databases\n$ poetry run preql -f dev/prepare_db.pql snowflake://<uri>\n$ poetry run preql -f dev/prepare_db.pql mssql://<uri>\n$ poetry run preql -f dev/prepare_db.pql bigquery:///<project>\n```\n\n**5. Run **data-diff** against seeded database (optional)**\n\n```bash\npoetry run python3 -m data_diff postgresql://postgres:Password1@localhost/postgres rating postgresql://postgres:Password1@localhost/postgres rating_del1 --verbose\n```\n\n**6. Run benchmarks (optional)**\n\n```shell-session\n$ dev/benchmark.sh #  runs benchmarks and puts results in benchmark_<sha>.csv\n$ poetry run python3 dev/graph.py #  create graphs from benchmark_*.csv files\n```\n\nYou can adjust how many rows we benchmark with by passing `N_SAMPLES` to `dev/benchmark.sh`:\n\n```shell-session\n$ N_SAMPLES=100000000 dev/benchmark.sh #  100m which is our canonical target\n```\n\n\n# License\n\n[MIT License](https://github.com/datafold/data-diff/blob/master/LICENSE)\n\n[dbs]: #supported-databases\n[tech-explain]: #technical-explanation\n[perf]: #performance-considerations\n[slack]: https://locallyoptimistic.com/community/\n',
    'author': 'Datafold',
    'author_email': 'data-diff@datafold.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/datafold/data-diff',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



#
# This module contains all the functionality related to the anonymous tracking of data-diff use.
#

import logging
import os
import json
import platform
from time import time
from typing import Any, Dict, Optional
import urllib.request
from uuid import uuid4

import toml

TRACK_URL = "https://api.perfalytics.com/track"
START_EVENT = "os_diff_run_start"
END_EVENT = "os_diff_run_end"
TOKEN = "ccb8c3a6-3b6f-445c-ad67-994efa7bd020"
TIMEOUT = 8

DEFAULT_PROFILE = os.path.expanduser("~/.datadiff.toml")


def _load_profile():
    try:
        with open(DEFAULT_PROFILE) as f:
            conf = toml.load(f)
    except FileNotFoundError:
        conf = {}

    if "anonymous_id" not in conf:
        conf["anonymous_id"] = str(uuid4())
        with open(DEFAULT_PROFILE, "w") as f:
            toml.dump(conf, f)
    return conf


g_tracking_enabled = True
g_anonymous_id = None


def disable_tracking():
    global g_tracking_enabled
    g_tracking_enabled = False


def is_tracking_enabled():
    return g_tracking_enabled


def get_anonymous_id():
    global g_anonymous_id
    if g_anonymous_id is None:
        profile = _load_profile()
        g_anonymous_id = profile["anonymous_id"]
    return g_anonymous_id


def create_start_event_json(diff_options: Dict[str, Any]):
    return {
        "event": "os_diff_run_start",
        "properties": {
            "distinct_id": get_anonymous_id(),
            "token": TOKEN,
            "time": time(),
            "os_type": os.name,
            "os_version": platform.platform(),
            "python_version": f"{platform.python_version()}/{platform.python_implementation()}",
            "diff_options": diff_options,
        },
    }


def create_end_event_json(
    is_success: bool,
    runtime_seconds: float,
    db1: str,
    db2: str,
    table1_count: int,
    table2_count: int,
    diff_count: int,
    error: Optional[str],
):
    return {
        "event": "os_diff_run_end",
        "properties": {
            "distinct_id": get_anonymous_id(),
            "token": TOKEN,
            "time": time(),
            "is_success": is_success,
            "runtime_seconds": runtime_seconds,
            "data_source_1_type": db1,
            "data_source_2_type": db2,
            "table_1_rows_cnt": table1_count,
            "table_2_rows_cnt": table2_count,
            "diff_rows_cnt": diff_count,
            "error_message": error,
        },
    }


def send_event_json(event_json):
    if not g_tracking_enabled:
        raise RuntimeError("Won't send; tracking is disabled!")

    headers = {"Content-Type": "application/json"}
    data = json.dumps(event_json).encode()
    try:
        req = urllib.request.Request(TRACK_URL, data=data, headers=headers)
        with urllib.request.urlopen(req, timeout=TIMEOUT) as f:
            res = f.read()
            if f.code != 200:
                raise RuntimeError(res)
    except Exception as e:
        logging.debug(f"Failed to post to freshpaint: {e}")



from runtype import dataclass

SKIP = object()


def _drop_skips(exprs):
    return [e for e in exprs if e is not SKIP]


class ITable:
    def __init__(self, *path):
        assert all(isinstance(str, i) for i in path)
        self.path = path

    def select(self, *exprs):
        exprs = _drop_skips(exprs)
        ...

    def where(self, *exprs):
        exprs = _drop_skips(exprs)
        if not exprs:
            return self

        ...

    def at(self, *exprs):
        exprs = _drop_skips(exprs)
        if not exprs:
            return self
        ...

    def join(self, target):
        return Join(self, target)

    def group_by(self, *, keys=None, values=None):
        assert keys or values
        pass

    def with_schema(self):
        pass

    def __getattr__(self, column):
        if column not in self.columns:
            raise KeyError()
        return Column(self, column)


class Table(ITable):
    def insert_values(self, rows):
        pass

    def insert_query(self, query):
        pass


class Q:
    def __getattr__(self, name):
        return _ResolveName(name)

    def __getitem__(self, name):
        return _ResolveName(name)


class Join(ITable):
    def on(self, *exprs):
        exprs = _drop_skips(exprs)
        if not exprs:
            return self


def join(*tables):
    "Joins each table into a 'struct'"
    ...


class Select(ITable):
    pass


class GroupBy(ITable):
    def having(self):
        pass


@dataclass
class Select(ITable):
    columns: Sequence[SqlOrStr]
    table: SqlOrStr = None
    where: Sequence[SqlOrStr] = None
    order_by: Sequence[SqlOrStr] = None
    group_by: Sequence[SqlOrStr] = None
    limit: int = None

    def compile(self, parent_c: Compiler):
        c = parent_c.replace(in_select=True)
        columns = ", ".join(map(c.compile, self.columns))
        select = f"SELECT {columns}"

        if self.table:
            select += " FROM " + c.compile(self.table)

        if self.where:
            select += " WHERE " + " AND ".join(map(c.compile, self.where))

        if self.group_by:
            select += " GROUP BY " + ", ".join(map(c.compile, self.group_by))

        if self.order_by:
            select += " ORDER BY " + ", ".join(map(c.compile, self.order_by))

        if self.limit is not None:
            select += " " + c.database.offset_limit(0, self.limit)

        if parent_c.in_select:
            select = "(%s)" % select
        return select


@dataclass
class Cte(ITable):
    expr: Expr
    name: str
    params: Tuple[str, ...]


def make_cte(t, *, name=None, params=None):
    return Cte()


def test1():
    src = Table("src_table")
    q = src.select(src.a, src.b).where(src.a > src.b + 1)



import re
import math
from typing import Iterable, Tuple, Union, Any, Sequence, Dict
from typing import TypeVar, Generic
from abc import ABC, abstractmethod
from urllib.parse import urlparse
from uuid import UUID
import operator
import string
import threading

alphanums = string.digits + string.ascii_lowercase


def safezip(*args):
    "zip but makes sure all sequences are the same length"
    assert len(set(map(len, args))) == 1
    return zip(*args)


def split_space(start, end, count):
    size = end - start
    assert count <= size, (count, size)
    return list(range(start, end, (size + 1) // (count + 1)))[1 : count + 1]


class ArithString:
    @classmethod
    def new(cls, *args, **kw):
        return cls(*args, **kw)


class ArithUUID(UUID, ArithString):
    "A UUID that supports basic arithmetic (add, sub)"

    def __int__(self):
        return self.int

    def __add__(self, other: Union[UUID, int]):
        if isinstance(other, int):
            return self.new(int=self.int + other)
        return NotImplemented

    def __sub__(self, other: Union[UUID, int]):
        if isinstance(other, int):
            return self.new(int=self.int - other)
        elif isinstance(other, UUID):
            return self.int - other.int
        return NotImplemented


def numberToBase(num, base):
    digits = []
    while num > 0:
        num, remainder = divmod(num, base)
        digits.append(remainder)
    return "".join(alphanums[i] for i in digits[::-1])


class ArithAlphanumeric(ArithString):
    def __init__(self, str: str = None, int: int = None, max_len=None):
        if str is None:
            str = numberToBase(int, len(alphanums))
        else:
            assert int is None

        if max_len and len(str) > max_len:
            raise ValueError(f"Length of alphanum value '{str}' is longer than the expected {max_len}")

        self._str = str
        self._max_len = max_len

    @property
    def int(self):
        return int(self._str, len(alphanums))

    def __str__(self):
        s = self._str
        if self._max_len:
            s = s.rjust(self._max_len, "0")
        return s

    def __len__(self):
        return len(self._str)

    def __int__(self):
        return self.int

    def __repr__(self):
        return f'alphanum"{self._str}"'

    def __add__(self, other: "Union[ArithAlphanumeric, int]"):
        if isinstance(other, int):
            res = self.new(int=self.int + other)
            if len(str(res)) != len(self):
                raise ValueError("Overflow error when adding to alphanumeric")
            return res
        return NotImplemented

    def __sub__(self, other: "Union[ArithAlphanumeric, int]"):
        if isinstance(other, int):
            return type(self)(int=self.int - other)
        elif isinstance(other, ArithAlphanumeric):
            return self.int - other.int
        return NotImplemented

    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.int >= other.int

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.int < other.int

    def new(self, *args, **kw):
        return type(self)(*args, **kw, max_len=self._max_len)


def is_uuid(u):
    try:
        UUID(u)
    except ValueError:
        return False
    return True


def number_to_human(n):
    millnames = ["", "k", "m", "b"]
    n = float(n)
    millidx = max(
        0,
        min(len(millnames) - 1, int(math.floor(0 if n == 0 else math.log10(abs(n)) / 3))),
    )

    return "{:.0f}{}".format(n / 10 ** (3 * millidx), millnames[millidx])


def _join_if_any(sym, args):
    args = list(args)
    if not args:
        return ""
    return sym.join(str(a) for a in args if a)


def remove_password_from_url(url: str, replace_with: str = "***") -> str:
    parsed = urlparse(url)
    account = parsed.username or ""
    if parsed.password:
        account += ":" + replace_with
    host = _join_if_any(":", filter(None, [parsed.hostname, parsed.port]))
    netloc = _join_if_any("@", filter(None, [account, host]))
    replaced = parsed._replace(netloc=netloc)
    return replaced.geturl()


def join_iter(joiner: Any, iterable: Iterable) -> Iterable:
    it = iter(iterable)
    yield next(it)
    for i in it:
        yield joiner
        yield i


V = TypeVar("V")


class CaseAwareMapping(ABC, Generic[V]):
    @abstractmethod
    def get_key(self, key: str) -> str:
        ...

    @abstractmethod
    def __getitem__(self, key: str) -> V:
        ...

    @abstractmethod
    def __setitem__(self, key: str, value: V):
        ...

    @abstractmethod
    def __contains__(self, key: str) -> bool:
        ...


class CaseInsensitiveDict(CaseAwareMapping):
    def __init__(self, initial):
        self._dict = {k.lower(): (k, v) for k, v in dict(initial).items()}

    def get_key(self, key: str) -> str:
        return self._dict[key.lower()][0]

    def __getitem__(self, key: str) -> V:
        return self._dict[key.lower()][1]

    def __setitem__(self, key: str, value):
        k = key.lower()
        if k in self._dict:
            key = self._dict[k][0]
        self._dict[k] = key, value

    def __contains__(self, key):
        return key.lower() in self._dict

    def keys(self) -> Iterable[str]:
        return self._dict.keys()

    def items(self) -> Iterable[Tuple[str, V]]:
        return ((k, v[1]) for k, v in self._dict.items())


class CaseSensitiveDict(dict, CaseAwareMapping):
    def get_key(self, key):
        return key

    def as_insensitive(self):
        return CaseInsensitiveDict(self)


def match_like(pattern: str, strs: Sequence[str]) -> Iterable[str]:
    reo = re.compile(pattern.replace("%", ".*").replace("?", ".") + "$")
    for s in strs:
        if reo.match(s):
            yield s


def accumulate(iterable, func=operator.add, *, initial=None):
    "Return running totals"
    # Taken from https://docs.python.org/3/library/itertools.html#itertools.accumulate, to backport 'initial' to 3.7
    it = iter(iterable)
    total = initial
    if initial is None:
        try:
            total = next(it)
        except StopIteration:
            return
    yield total
    for element in it:
        total = func(total, element)
        yield total


def match_regexps(regexps: Dict[str, Any], s: str) -> Sequence[tuple]:
    for regexp, v in regexps.items():
        m = re.match(regexp + "$", s)
        if m:
            yield m, v


def run_as_daemon(threadfunc, *args):
    th = threading.Thread(target=threadfunc, args=args)
    th.daemon = True
    th.start()
    return th



"""Provides classes for performing a table diff
"""

import time
import os
from numbers import Number
from operator import attrgetter, methodcaller
from collections import defaultdict
from typing import List, Tuple, Iterator, Optional
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed

from runtype import dataclass

from .tracking import create_end_event_json, create_start_event_json, send_event_json, is_tracking_enabled
from .sql import Select, Checksum, Compare, Count, TableName, Time, Value
from .utils import (
    CaseAwareMapping,
    CaseInsensitiveDict,
    safezip,
    split_space,
    CaseSensitiveDict,
    ArithString,
    run_as_daemon,
)
from .databases.base import Database
from .databases.database_types import (
    DbPath,
    DbKey,
    DbTime,
    IKey,
    Native_UUID,
    NumericType,
    PrecisionType,
    StringType,
    Schema,
)

logger = logging.getLogger("diff_tables")

RECOMMENDED_CHECKSUM_DURATION = 10
BENCHMARK = os.environ.get("BENCHMARK", False)
DEFAULT_BISECTION_THRESHOLD = 1024 * 16
DEFAULT_BISECTION_FACTOR = 32


def create_schema(db: Database, table_path: DbPath, schema: dict, case_sensitive: bool) -> CaseAwareMapping:
    logger.debug(f"[{db.name}] Schema = {schema}")

    if case_sensitive:
        return CaseSensitiveDict(schema)

    if len({k.lower() for k in schema}) < len(schema):
        logger.warning(f'Ambiguous schema for {db}:{".".join(table_path)} | Columns = {", ".join(list(schema))}')
        logger.warning("We recommend to disable case-insensitivity (remove --any-case).")
    return CaseInsensitiveDict(schema)


@dataclass(frozen=False)
class TableSegment:
    """Signifies a segment of rows (and selected columns) within a table

    Parameters:
        database (Database): Database instance. See :meth:`connect`
        table_path (:data:`DbPath`): Path to table in form of a tuple. e.g. `('my_dataset', 'table_name')`
        key_column (str): Name of the key column, which uniquely identifies each row (usually id)
        update_column (str, optional): Name of updated column, which signals that rows changed (usually updated_at or last_update)
        extra_columns (Tuple[str, ...], optional): Extra columns to compare
        min_key (:data:`DbKey`, optional): Lowest key_column value, used to restrict the segment
        max_key (:data:`DbKey`, optional): Highest key_column value, used to restrict the segment
        min_update (:data:`DbTime`, optional): Lowest update_column value, used to restrict the segment
        max_update (:data:`DbTime`, optional): Highest update_column value, used to restrict the segment
        where (str, optional): An additional 'where' expression to restrict the search space.

        case_sensitive (bool): If false, the case of column names will adjust according to the schema. Default is true.

    """

    # Location of table
    database: Database
    table_path: DbPath

    # Columns
    key_column: str
    update_column: str = None
    extra_columns: Tuple[str, ...] = ()

    # Restrict the segment
    min_key: DbKey = None
    max_key: DbKey = None
    min_update: DbTime = None
    max_update: DbTime = None

    where: str = None
    case_sensitive: bool = True
    _schema: Schema = None

    def __post_init__(self):
        if not self.update_column and (self.min_update or self.max_update):
            raise ValueError("Error: the min_update/max_update feature requires 'update_column' to be set.")

        if self.min_key is not None and self.max_key is not None and self.min_key >= self.max_key:
            raise ValueError(f"Error: min_key expected to be smaller than max_key! ({self.min_key} >= {self.max_key})")

        if self.min_update is not None and self.max_update is not None and self.min_update >= self.max_update:
            raise ValueError(
                f"Error: min_update expected to be smaller than max_update! ({self.min_update} >= {self.max_update})"
            )

    @property
    def _update_column(self):
        return self._quote_column(self.update_column)

    def _quote_column(self, c: str) -> str:
        if self._schema:
            c = self._schema.get_key(c)  # Get the actual name. Might be case-insensitive.
        return self.database.quote(c)

    def _normalize_column(self, name: str, template: str = None) -> str:
        if not self._schema:
            raise RuntimeError(
                "Cannot compile query when the schema is unknown. Please use TableSegment.with_schema()."
            )

        col_type = self._schema[name]
        col = self._quote_column(name)

        if isinstance(col_type, Native_UUID):
            # Normalize first, apply template after (for uuids)
            # Needed because min/max(uuid) fails in postgresql
            col = self.database.normalize_value_by_type(col, col_type)
            if template is not None:
                col = template % col  # Apply template using Python's string formatting
            return col

        # Apply template before normalizing (for ints)
        if template is not None:
            col = template % col  # Apply template using Python's string formatting

        return self.database.normalize_value_by_type(col, col_type)

    def _with_raw_schema(self, raw_schema: dict) -> "TableSegment":
        schema = self.database._process_table_schema(self.table_path, raw_schema, self._relevant_columns)
        return self.new(_schema=create_schema(self.database, self.table_path, schema, self.case_sensitive))

    def with_schema(self) -> "TableSegment":
        "Queries the table schema from the database, and returns a new instance of TableSegment, with a schema."
        if self._schema:
            return self

        return self._with_raw_schema(self.database.query_table_schema(self.table_path))

    def _make_key_range(self):
        if self.min_key is not None:
            yield Compare("<=", Value(self.min_key), self._quote_column(self.key_column))
        if self.max_key is not None:
            yield Compare("<", self._quote_column(self.key_column), Value(self.max_key))

    def _make_update_range(self):
        if self.min_update is not None:
            yield Compare("<=", Time(self.min_update), self._update_column)
        if self.max_update is not None:
            yield Compare("<", self._update_column, Time(self.max_update))

    def _make_select(self, *, table=None, columns=None, where=None, group_by=None, order_by=None):
        if columns is None:
            columns = [self._normalize_column(self.key_column)]
        where = [
            *self._make_key_range(),
            *self._make_update_range(),
            *([] if where is None else [where]),
            *([] if self.where is None else [self.where]),
        ]
        order_by = None if order_by is None else [order_by]
        return Select(
            table=table or TableName(self.table_path),
            where=where,
            columns=columns,
            group_by=group_by,
            order_by=order_by,
        )

    def get_values(self) -> list:
        "Download all the relevant values of the segment from the database"
        select = self._make_select(columns=self._relevant_columns_repr)
        return self.database.query(select, List[Tuple])

    def choose_checkpoints(self, count: int) -> List[DbKey]:
        "Suggests a bunch of evenly-spaced checkpoints to split by (not including start, end)"
        assert self.is_bounded
        if isinstance(self.min_key, ArithString):
            assert type(self.min_key) is type(self.max_key)
            checkpoints = split_space(self.min_key.int, self.max_key.int, count)
            return [self.min_key.new(int=i) for i in checkpoints]

        return split_space(self.min_key, self.max_key, count)

    def segment_by_checkpoints(self, checkpoints: List[DbKey]) -> List["TableSegment"]:
        "Split the current TableSegment to a bunch of smaller ones, separated by the given checkpoints"

        if self.min_key and self.max_key:
            assert all(self.min_key <= c < self.max_key for c in checkpoints)
        checkpoints.sort()

        # Calculate sub-segments
        positions = [self.min_key] + checkpoints + [self.max_key]
        ranges = list(zip(positions[:-1], positions[1:]))

        # Create table segments
        tables = [self.new(min_key=s, max_key=e) for s, e in ranges]

        return tables

    def new(self, **kwargs) -> "TableSegment":
        """Using new() creates a copy of the instance using 'replace()'"""
        return self.replace(**kwargs)

    @property
    def _relevant_columns(self) -> List[str]:
        extras = list(self.extra_columns)

        if self.update_column and self.update_column not in extras:
            extras = [self.update_column] + extras

        return [self.key_column] + extras

    @property
    def _relevant_columns_repr(self) -> List[str]:
        return [self._normalize_column(c) for c in self._relevant_columns]

    def count(self) -> Tuple[int, int]:
        """Count how many rows are in the segment, in one pass."""
        return self.database.query(self._make_select(columns=[Count()]), int)

    def count_and_checksum(self) -> Tuple[int, int]:
        """Count and checksum the rows in the segment, in one pass."""
        start = time.monotonic()
        count, checksum = self.database.query(
            self._make_select(columns=[Count(), Checksum(self._relevant_columns_repr)]), tuple
        )
        duration = time.monotonic() - start
        if duration > RECOMMENDED_CHECKSUM_DURATION:
            logger.warning(
                f"Checksum is taking longer than expected ({duration:.2f}s). "
                "We recommend increasing --bisection-factor or decreasing --threads."
            )

        if count:
            assert checksum, (count, checksum)
        return count or 0, checksum if checksum is None else int(checksum)

    def query_key_range(self) -> Tuple[int, int]:
        """Query database for minimum and maximum key. This is used for setting the initial bounds."""
        # Normalizes the result (needed for UUIDs) after the min/max computation
        select = self._make_select(
            columns=[
                self._normalize_column(self.key_column, "min(%s)"),
                self._normalize_column(self.key_column, "max(%s)"),
            ]
        )
        min_key, max_key = self.database.query(select, tuple)

        if min_key is None or max_key is None:
            raise ValueError("Table appears to be empty")

        return min_key, max_key

    @property
    def is_bounded(self):
        return self.min_key is not None and self.max_key is not None

    def approximate_size(self):
        if not self.is_bounded:
            raise RuntimeError("Cannot approximate the size of an unbounded segment. Must have min_key and max_key.")
        return self.max_key - self.min_key


def diff_sets(a: set, b: set) -> Iterator:
    s1 = set(a)
    s2 = set(b)
    d = defaultdict(list)

    # The first item is always the key (see TableDiffer._relevant_columns)
    for i in s1 - s2:
        d[i[0]].append(("-", i))
    for i in s2 - s1:
        d[i[0]].append(("+", i))

    for _k, v in sorted(d.items(), key=lambda i: i[0]):
        yield from v


DiffResult = Iterator[Tuple[str, tuple]]  # Iterator[Tuple[Literal["+", "-"], tuple]]


@dataclass
class TableDiffer:
    """Finds the diff between two SQL tables

    The algorithm uses hashing to quickly check if the tables are different, and then applies a
    bisection search recursively to find the differences efficiently.

    Works best for comparing tables that are mostly the same, with minor discrepencies.

    Parameters:
        bisection_factor (int): Into how many segments to bisect per iteration.
        bisection_threshold (Number): When should we stop bisecting and compare locally (in row count).
        threaded (bool): Enable/disable threaded diffing. Needed to take advantage of database threads.
        max_threadpool_size (int): Maximum size of each threadpool. ``None`` means auto. Only relevant when `threaded` is ``True``.
                                   There may be many pools, so number of actual threads can be a lot higher.
    """

    bisection_factor: int = DEFAULT_BISECTION_FACTOR
    bisection_threshold: Number = DEFAULT_BISECTION_THRESHOLD  # Accepts inf for tests
    threaded: bool = True
    max_threadpool_size: Optional[int] = 1

    # Enable/disable debug prints
    debug: bool = False

    stats: dict = {}

    def diff_tables(self, table1: TableSegment, table2: TableSegment) -> DiffResult:
        """Diff the given tables.

        Parameters:
            table1 (TableSegment): The "before" table to compare. Or: source table
            table2 (TableSegment): The "after" table to compare. Or: target table

        Returns:
            An iterator that yield pair-tuples, representing the diff. Items can be either
            ('-', columns) for items in table1 but not in table2
            ('+', columns) for items in table2 but not in table1
            Where `columns` is a tuple of values for the involved columns, i.e. (id, ...extra)
        """
        # Validate options
        if self.bisection_factor >= self.bisection_threshold:
            raise ValueError("Incorrect param values (bisection factor must be lower than threshold)")
        if self.bisection_factor < 2:
            raise ValueError("Must have at least two segments per iteration (i.e. bisection_factor >= 2)")

        if is_tracking_enabled():
            options = dict(self)
            event_json = create_start_event_json(options)
            run_as_daemon(send_event_json, event_json)

        self.stats["diff_count"] = 0
        start = time.monotonic()
        try:

            # Query and validate schema
            table1, table2 = self._threaded_call("with_schema", [table1, table2])
            self._validate_and_adjust_columns(table1, table2)

            key_type = table1._schema[table1.key_column]
            key_type2 = table2._schema[table2.key_column]
            if not isinstance(key_type, IKey):
                raise NotImplementedError(f"Cannot use column of type {key_type} as a key")
            if not isinstance(key_type2, IKey):
                raise NotImplementedError(f"Cannot use column of type {key_type2} as a key")
            assert key_type.python_type is key_type2.python_type

            # Query min/max values
            key_ranges = self._threaded_call_as_completed("query_key_range", [table1, table2])

            # Start with the first completed value, so we don't waste time waiting
            min_key1, max_key1 = self._parse_key_range_result(key_type, next(key_ranges))

            table1, table2 = [t.new(min_key=min_key1, max_key=max_key1) for t in (table1, table2)]

            logger.info(
                f"Diffing tables | segments: {self.bisection_factor}, bisection threshold: {self.bisection_threshold}. "
                f"key-range: {table1.min_key}..{table2.max_key}, "
                f"size: {table1.approximate_size()}"
            )

            # Bisect (split) the table into segments, and diff them recursively.
            yield from self._bisect_and_diff_tables(table1, table2)

            # Now we check for the second min-max, to diff the portions we "missed".
            min_key2, max_key2 = self._parse_key_range_result(key_type, next(key_ranges))

            if min_key2 < min_key1:
                pre_tables = [t.new(min_key=min_key2, max_key=min_key1) for t in (table1, table2)]
                yield from self._bisect_and_diff_tables(*pre_tables)

            if max_key2 > max_key1:
                post_tables = [t.new(min_key=max_key1, max_key=max_key2) for t in (table1, table2)]
                yield from self._bisect_and_diff_tables(*post_tables)

            error = None
        except BaseException as e:  # Catch KeyboardInterrupt too
            error = e
        finally:
            if is_tracking_enabled():
                runtime = time.monotonic() - start
                table1_count = self.stats.get("table1_count")
                table2_count = self.stats.get("table2_count")
                diff_count = self.stats.get("diff_count")
                err_message = str(error)[:20]  # Truncate possibly sensitive information.
                event_json = create_end_event_json(
                    error is None,
                    runtime,
                    table1.database.name,
                    table2.database.name,
                    table1_count,
                    table2_count,
                    diff_count,
                    err_message,
                )
                send_event_json(event_json)

            if error:
                raise error

    def _parse_key_range_result(self, key_type, key_range):
        mn, mx = key_range
        cls = key_type.make_value
        # We add 1 because our ranges are exclusive of the end (like in Python)
        try:
            return cls(mn), cls(mx) + 1
        except (TypeError, ValueError) as e:
            raise type(e)(f"Cannot apply {key_type} to {mn}, {mx}.") from e

    def _validate_and_adjust_columns(self, table1, table2):
        for c in table1._relevant_columns:
            if c not in table1._schema:
                raise ValueError(f"Column '{c}' not found in schema for table {table1}")
            if c not in table2._schema:
                raise ValueError(f"Column '{c}' not found in schema for table {table2}")

            # Update schemas to minimal mutual precision
            col1 = table1._schema[c]
            col2 = table2._schema[c]
            if isinstance(col1, PrecisionType):
                if not isinstance(col2, PrecisionType):
                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")

                lowest = min(col1, col2, key=attrgetter("precision"))

                if col1.precision != col2.precision:
                    logger.warning(f"Using reduced precision {lowest} for column '{c}'. Types={col1}, {col2}")

                table1._schema[c] = col1.replace(precision=lowest.precision, rounds=lowest.rounds)
                table2._schema[c] = col2.replace(precision=lowest.precision, rounds=lowest.rounds)

            elif isinstance(col1, NumericType):
                if not isinstance(col2, NumericType):
                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")

                lowest = min(col1, col2, key=attrgetter("precision"))

                if col1.precision != col2.precision:
                    logger.warning(f"Using reduced precision {lowest} for column '{c}'. Types={col1}, {col2}")

                table1._schema[c] = col1.replace(precision=lowest.precision)
                table2._schema[c] = col2.replace(precision=lowest.precision)

            elif isinstance(col1, StringType):
                if not isinstance(col2, StringType):
                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")

        for t in [table1, table2]:
            for c in t._relevant_columns:
                ctype = t._schema[c]
                if not ctype.supported:
                    logger.warning(
                        f"[{t.database.name}] Column '{c}' of type '{ctype}' has no compatibility handling. "
                        "If encoding/formatting differs between databases, it may result in false positives."
                    )

    def _bisect_and_diff_tables(self, table1, table2, level=0, max_rows=None):
        assert table1.is_bounded and table2.is_bounded

        if max_rows is None:
            # We can be sure that row_count <= max_rows
            max_rows = table1.max_key - table1.min_key

        # If count is below the threshold, just download and compare the columns locally
        # This saves time, as bisection speed is limited by ping and query performance.
        if max_rows < self.bisection_threshold:
            rows1, rows2 = self._threaded_call("get_values", [table1, table2])
            diff = list(diff_sets(rows1, rows2))

            # Initial bisection_threshold larger than count. Normally we always
            # checksum and count segments, even if we get the values. At the
            # first level, however, that won't be true.
            if level == 0:
                self.stats["table1_count"] = len(rows1)
                self.stats["table2_count"] = len(rows2)

            self.stats["diff_count"] += len(diff)

            logger.info(". " * level + f"Diff found {len(diff)} different rows.")
            self.stats["rows_downloaded"] = self.stats.get("rows_downloaded", 0) + max(len(rows1), len(rows2))
            yield from diff
            return

        # Choose evenly spaced checkpoints (according to min_key and max_key)
        checkpoints = table1.choose_checkpoints(self.bisection_factor - 1)

        # Create new instances of TableSegment between each checkpoint
        segmented1 = table1.segment_by_checkpoints(checkpoints)
        segmented2 = table2.segment_by_checkpoints(checkpoints)

        # Recursively compare each pair of corresponding segments between table1 and table2
        diff_iters = [
            self._diff_tables(t1, t2, level + 1, i + 1, len(segmented1))
            for i, (t1, t2) in enumerate(safezip(segmented1, segmented2))
        ]

        for res in self._thread_map(list, diff_iters):
            yield from res

    def _diff_tables(self, table1, table2, level=0, segment_index=None, segment_count=None):
        logger.info(
            ". " * level + f"Diffing segment {segment_index}/{segment_count}, "
            f"key-range: {table1.min_key}..{table2.max_key}, "
            f"size: {table2.max_key-table1.min_key}"
        )

        # When benchmarking, we want the ability to skip checksumming. This
        # allows us to download all rows for comparison in performance. By
        # default, data-diff will checksum the section first (when it's below
        # the threshold) and _then_ download it.
        if BENCHMARK:
            max_rows_from_keys = max(table1.max_key - table1.min_key, table2.max_key - table2.min_key)
            if max_rows_from_keys < self.bisection_threshold:
                yield from self._bisect_and_diff_tables(table1, table2, level=level, max_rows=max_rows_from_keys)
                return

        (count1, checksum1), (count2, checksum2) = self._threaded_call("count_and_checksum", [table1, table2])

        if count1 == 0 and count2 == 0:
            logger.warning(
                "Uneven distribution of keys detected. (big gaps in the key column). "
                "For better performance, we recommend to increase the bisection-threshold."
            )
            assert checksum1 is None and checksum2 is None
            return

        if level == 1:
            self.stats["table1_count"] = self.stats.get("table1_count", 0) + count1
            self.stats["table2_count"] = self.stats.get("table2_count", 0) + count2

        if checksum1 != checksum2:
            yield from self._bisect_and_diff_tables(table1, table2, level=level, max_rows=max(count1, count2))

    def _thread_map(self, func, iterable):
        if not self.threaded:
            return map(func, iterable)

        with ThreadPoolExecutor(max_workers=self.max_threadpool_size) as task_pool:
            return task_pool.map(func, iterable)

    def _threaded_call(self, func, iterable):
        "Calls a method for each object in iterable."
        return list(self._thread_map(methodcaller(func), iterable))

    def _thread_as_completed(self, func, iterable):
        if not self.threaded:
            return map(func, iterable)

        with ThreadPoolExecutor(max_workers=self.max_threadpool_size) as task_pool:
            futures = [task_pool.submit(func, item) for item in iterable]
            for future in as_completed(futures):
                yield future.result()

    def _threaded_call_as_completed(self, func, iterable):
        "Calls a method for each object in iterable. Returned in order of completion."
        return self._thread_as_completed(methodcaller(func), iterable)
